""".. Ignore pydocstyle D400.

=================
Clear Redis Cache
=================

Clear all the Redis Cache.
"""

from django.core.management.base import BaseCommand

from resolwe.flow.managers.listener.redis_cache import redis_cache


class Command(BaseCommand):
    """Run the executor listener."""

    help = "Clear Redis cache."

    def handle(self, *args, **options):
        """Clear the Redis cache."""
        redis_cache.clear()



# Generated by Django 3.2.19 on 2023-06-13 19:17

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("flow", "0010_annotationfield_required"),
    ]

    operations = [
        migrations.AlterField(
            model_name="annotationpreset",
            name="contributor",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT, to=settings.AUTH_USER_MODEL
            ),
        ),
    ]



# pylint: disable=missing-docstring
import datetime

from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group
from django.utils.timezone import get_current_timezone
from rest_framework import status
from rest_framework.test import APIRequestFactory, force_authenticate

from resolwe.flow.models import (
    Collection,
    Data,
    DescriptorSchema,
    Entity,
    Process,
    Relation,
    RelationType,
)
from resolwe.flow.views import (
    CollectionViewSet,
    DataViewSet,
    DescriptorSchemaViewSet,
    EntityViewSet,
    ProcessViewSet,
    RelationViewSet,
)
from resolwe.permissions.models import Permission, get_anonymous_user
from resolwe.test import TestCase

factory = APIRequestFactory()


class BaseViewSetFiltersTest(TestCase):
    @classmethod
    def setUpTestData(cls):
        user_model = get_user_model()
        cls.admin = user_model.objects.create_superuser(
            username="admin",
            email="admin@test.com",
            password="admin",
            first_name="James",
            last_name="Smith",
        )
        cls.contributor = user_model.objects.create_user(
            username="contributor",
            email="contributor@test.com",
            first_name="Joe",
            last_name="Miller",
        )
        cls.user = user_model.objects.create_user(
            username="normal_user",
            email="user@test.com",
            first_name="John",
            last_name="Williams",
        )
        cls.group = Group.objects.create(name="Test group")

    def setUp(self):
        self.viewset = self.viewset_class.as_view(
            actions={
                "get": "list",
            }
        )

    def _check_filter(
        self, query_args, expected, expected_status_code=status.HTTP_200_OK, user=None
    ):
        """Check that query_args filter to expected queryset."""
        request = factory.get("/", query_args, format="json")
        force_authenticate(request, user or self.admin)
        response = self.viewset(request)

        if status.is_success(response.status_code):
            self.assertEqual(len(response.data), len(expected))
            self.assertCountEqual(
                [item.pk for item in expected],
                [item["id"] for item in response.data],
            )
        else:
            self.assertEqual(response.status_code, expected_status_code)
            response.render()

        return response


class CollectionViewSetFiltersTest(BaseViewSetFiltersTest):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()

        cls.descriptor_schema = DescriptorSchema.objects.create(
            slug="test-schema",
            version="1.0.0",
            contributor=cls.contributor,
            schema=[
                {
                    "name": "company",
                    "group": [
                        {
                            "name": "name",
                            "type": "basic:string:",
                            "required": False,
                        },
                        {
                            "name": "departments",
                            "type": "list:basic:string:",
                            "required": False,
                        },
                    ],
                }
            ],
        )

        cls.collections = [
            Collection.objects.create(
                name="Test collection 0",
                slug="test-collection-0",
                contributor=cls.contributor,
                description="My description!",
                descriptor_schema=cls.descriptor_schema,
                descriptor={
                    "company": {
                        "name": "Genialis",
                        "departments": ["engineering", "operations"],
                    }
                },
                tags=["first-tag"],
            ),
            Collection.objects.create(
                name="Test collection 1",
                slug="test-collection-1",
                contributor=cls.contributor,
                description="My favourite test collection",
                descriptor_schema=cls.descriptor_schema,
                tags=["first-tag", "second-tag"],
            ),
            Collection.objects.create(
                name="User test collection",
                slug="user-collection",
                contributor=cls.user,
                description="User's description",
                tags=["second-tag"],
            ),
            Collection.objects.create(
                name="Public test collection",
                slug="public-collection",
                contributor=cls.admin,
            ),
        ]

        cls.entity1 = Entity.objects.create(
            contributor=cls.contributor, collection=cls.collections[1]
        )
        cls.entity2 = Entity.objects.create(
            contributor=cls.contributor, collection=cls.collections[2]
        )
        cls.entity3 = Entity.objects.create(
            contributor=cls.contributor, collection=cls.collections[2]
        )
        cls.entity1.save()
        cls.entity2.save()
        cls.entity3.save()

        tzone = get_current_timezone()
        cls.collections[0].created = datetime.datetime(2016, 7, 30, 0, 59, tzinfo=tzone)
        cls.collections[0].save()
        cls.collections[1].created = datetime.datetime(2016, 7, 30, 1, 59, tzinfo=tzone)
        cls.collections[1].save()
        cls.collections[2].created = datetime.datetime(2016, 7, 30, 2, 59, tzinfo=tzone)
        cls.collections[2].save()

        cls.collections[0].set_permission(Permission.OWNER, cls.admin)
        cls.collections[0].set_permission(Permission.VIEW, cls.user)
        cls.collections[1].set_permission(Permission.OWNER, cls.admin)
        cls.collections[1].set_permission(Permission.VIEW, cls.user)
        cls.collections[2].set_permission(Permission.OWNER, cls.user)
        cls.collections[2].set_permission(Permission.OWNER, cls.group)
        cls.collections[3].set_permission(Permission.VIEW, get_anonymous_user())

        cls.viewset_class = CollectionViewSet

    def test_filter_id(self):
        self._check_filter({"id": self.collections[0].pk}, [self.collections[0]])
        self._check_filter({"id": "420"}, [])
        self._check_filter(
            {"id__in": "{},{}".format(self.collections[0].pk, self.collections[1].pk)},
            self.collections[:2],
        )
        self._check_filter({"id__gt": self.collections[0].pk}, self.collections[1:])

    def test_filter_slug(self):
        self._check_filter({"slug": "test-collection-0"}, [self.collections[0]])
        self._check_filter(
            {"slug__in": "test-collection-1,user-collection"}, self.collections[1:3]
        )

    def test_filter_name(self):
        self._check_filter({"name": "Test collection 1"}, [self.collections[1]])
        self._check_filter({"name": "Test collection"}, [])

        self._check_filter({"name__iexact": "test collection 1"}, [self.collections[1]])

        self._check_filter(
            {"name__startswith": "Test collection"}, self.collections[:2]
        )
        self._check_filter(
            {"name__istartswith": "test collection"}, self.collections[:2]
        )

        self._check_filter({"name__contains": "test"}, self.collections[2:])
        self._check_filter({"name__icontains": "test"}, self.collections)

    def test_filter_description(self):
        self._check_filter(
            {"description__icontains": "Favourite"}, [self.collections[1]]
        )
        self._check_filter({"description__contains": "Favourite"}, [])
        self._check_filter({"description__contains": 420}, [])

    def test_filter_descriptor_schema(self):
        self._check_filter(
            {"descriptor_schema": str(self.descriptor_schema.pk)}, self.collections[:2]
        )
        self._check_filter({"descriptor_schema": "999999"}, [])

    def test_mixed_descriptor_schema_version(self):
        """Test processes with mixed descriptor schema versions are shown.

        When multiple data objects with different versions of the descriptor schema
        with same slug are in the filter result both process versions must be returned.
        """
        descriptor_schema = DescriptorSchema.objects.create(
            slug="test-schema",
            version="1.4.2",
            contributor=self.contributor,
            schema=[
                {
                    "name": "company",
                    "group": [
                        {
                            "name": "name",
                            "type": "basic:string:",
                            "required": False,
                        },
                        {
                            "name": "departments",
                            "type": "list:basic:string:",
                            "required": False,
                        },
                    ],
                }
            ],
        )
        Collection.objects.create(
            contributor=self.contributor, descriptor_schema=descriptor_schema
        ),
        request = factory.get("/", {}, format="json")
        force_authenticate(request, self.admin)
        response = self.viewset(request)
        self.assertSetEqual(
            {"1.0.0", "1.4.2"},
            {
                entry["descriptor_schema"]["version"]
                for entry in response.data
                if entry["descriptor_schema"] is not None
            },
        )

    def test_filter_tags(self):
        self._check_filter({"tags": "first-tag"}, self.collections[:2])
        self._check_filter({"tags": "first-tag,second-tag"}, [self.collections[1]])

    def test_filter_contributor(self):
        self._check_filter(
            {"contributor": str(self.contributor.pk)}, self.collections[:2]
        )
        self._check_filter(
            {"contributor__in": "{},{}".format(self.contributor.pk, self.user.pk)},
            self.collections[:3],
        )

    def test_filter_owners(self):
        self._check_filter({"owners": str(self.admin.pk)}, self.collections[:2])
        self._check_filter({"owners": "999999"}, [])

    def test_filter_contributor_name(self):
        # Filter by first name.
        self._check_filter({"contributor_name": "Joe"}, self.collections[:2])
        # Filter by last name.
        self._check_filter({"contributor_name": "Miller"}, self.collections[:2])
        # Filter by username.
        self._check_filter({"contributor_name": "contributor"}, self.collections[:2])
        # Filter by combination of first and last name.
        self._check_filter({"contributor_name": "John Williams"}, [self.collections[2]])

    def test_filter_owners_name(self):
        # Filter by first name.
        self._check_filter({"owners_name": "James"}, self.collections[:2])
        # Filter by last name.
        self._check_filter({"owners_name": "Smith"}, self.collections[:2])
        # Filter by username.
        self._check_filter({"owners_name": "admin"}, self.collections[:2])
        # Filter by combination of first and last name.
        self._check_filter({"owners_name": "John Williams"}, [self.collections[2]])

    def test_filter_permission(self):
        self._check_filter(
            {"permission": "edit"}, [self.collections[2]], user=self.user
        )
        self._check_filter(
            {"permission": "view"},
            self.collections,
            user=self.user,
        )
        self._check_filter(
            {"permission": "owner"}, [self.collections[2]], user=self.user
        )

    def test_filter_created(self):
        self._check_filter({"created": "2016-07-30T00:59:00"}, self.collections[:1])
        self._check_filter({"created__gt": "2016-07-30T01:59:00"}, self.collections[2:])
        self._check_filter(
            {"created__gte": "2016-07-30T01:59:00"}, self.collections[1:]
        )
        self._check_filter({"created__lt": "2016-07-30T01:59:00"}, self.collections[:1])
        self._check_filter(
            {"created__lte": "2016-07-30T01:59:00"}, self.collections[:2]
        )

    def test_filter_modified(self):
        self._check_filter(
            {"modified": self.collections[0].modified.isoformat()}, self.collections[:1]
        )
        self._check_filter(
            {"modified__gt": self.collections[1].modified.isoformat()},
            self.collections[2:3],
        )
        self._check_filter(
            {"modified__gte": self.collections[1].modified.isoformat()},
            self.collections[1:3],
        )
        self._check_filter(
            {"modified__lt": self.collections[1].modified.isoformat()},
            self.collections[:1] + [self.collections[3]],
        )
        self._check_filter(
            {"modified__lte": self.collections[1].modified.isoformat()},
            self.collections[:2] + [self.collections[3]],
        )

    def test_filter_text(self):
        # By name.
        self._check_filter({"text": "Test collection 1"}, [self.collections[1]])
        self._check_filter({"text": "Test"}, self.collections)
        self._check_filter({"text": "test"}, self.collections)
        self._check_filter({"text": "user"}, [self.collections[2]])

        # Test unaccented search.
        self._check_filter({"text": "Tešt čollećtion 1"}, [self.collections[1]])

        accented_collection = Collection.objects.create(
            name="Böhm", contributor=self.admin
        )
        accented_collection.set_permission(Permission.VIEW, self.user)
        self._check_filter({"text": "Bohm"}, [accented_collection])
        self._check_filter({"text": "Böhm"}, [accented_collection])
        accented_collection.delete()

        # By contributor.
        self._check_filter({"text": "joe"}, self.collections[:2])
        self._check_filter({"text": "Miller"}, self.collections[:2])

        # By owner.
        self._check_filter(
            {"text": "james"}, self.collections[:2] + [self.collections[3]]
        )
        self._check_filter(
            {"text": "Smith"}, self.collections[:2] + [self.collections[3]]
        )

        # By description.
        self._check_filter(
            {"text": "description"}, [self.collections[0], self.collections[2]]
        )
        self._check_filter({"text": "my"}, self.collections[:2])
        self._check_filter({"text": "my description"}, [self.collections[0]])
        self._check_filter({"text": "user"}, [self.collections[2]])

        # By descriptor.
        self._check_filter({"text": "genialis"}, [self.collections[0]])
        self._check_filter({"text": "engineering"}, [self.collections[0]])

        # By mixed fields.
        self._check_filter({"text": "test joe"}, self.collections[:2])
        self._check_filter({"text": "joe my description"}, [self.collections[0]])

        # Check that changes are applied immediately.
        self.collections[0].name = "Awesome new name"
        self.collections[0].save()

        self._check_filter({"text": "awesome"}, [self.collections[0]])

    def test_filter_mixed(self):
        self._check_filter(
            {"name": "Test collection 0", "owners_name": "James Smith"},
            [self.collections[0]],
        )
        self._check_filter(
            {"text": "collection 0", "owners_name": "James Smith"},
            [self.collections[0]],
        )

    def test_order_by_relevance(self):
        result = self._check_filter({"text": "test collection"}, self.collections)
        self.assertEqual(result.data[0]["id"], self.collections[1].pk)

        # Make another collection more important.
        collection = self.collections[2]
        collection.description = "This is test collection. My test collection rocks."
        collection.save()

        result = self._check_filter({"text": "test collection"}, self.collections)
        self.assertEqual(result.data[0]["id"], self.collections[2].pk)

        # Check that ordering can be overriden.
        result = self._check_filter(
            {"text": "test collection", "ordering": "id"}, self.collections
        )
        self.assertEqual(result.data[0]["id"], self.collections[0].pk)

    def test_filter_entity_count(self):
        self._check_filter(
            {"entity_count": 0}, [self.collections[0], self.collections[3]]
        )
        self._check_filter({"entity_count": 2}, [self.collections[2]])
        self._check_filter({"entity_count__gt": 1}, [self.collections[2]])
        self._check_filter(
            {"entity_count__lt": 1}, [self.collections[0], self.collections[3]]
        )
        self._check_filter(
            {"entity_count__gte": 1}, [self.collections[1], self.collections[2]]
        )
        self._check_filter(
            {"entity_count__lte": 1},
            [self.collections[0], self.collections[1], self.collections[3]],
        )

    def test_filter_group(self):
        self._check_filter({"group": self.group.pk}, [self.collections[2]])

    def test_filter_shared_with_me(self):
        self._check_filter(
            {"shared_with_me": True}, self.collections[:3], user=self.user
        )
        self._check_filter(
            {"shared_with_me": False}, [self.collections[3]], user=self.user
        )


class EntityViewSetFiltersTest(BaseViewSetFiltersTest):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()

        cls.descriptor_schema = DescriptorSchema.objects.create(
            slug="test-schema",
            version="1.0.0",
            contributor=cls.contributor,
            schema=[
                {
                    "name": "company",
                    "group": [
                        {
                            "name": "name",
                            "type": "basic:string:",
                            "required": False,
                        },
                        {
                            "name": "departments",
                            "type": "list:basic:string:",
                            "required": False,
                        },
                    ],
                }
            ],
        )

        cls.collection1 = Collection.objects.create(
            name="My collection",
            slug="my-collection",
            contributor=cls.contributor,
        )

        cls.collection2 = Collection.objects.create(
            name="Other collection",
            slug="other-collection",
            contributor=cls.contributor,
        )

        cls.entities = [
            Entity.objects.create(
                name="Test entity 0",
                slug="test-entity-0",
                contributor=cls.contributor,
                collection=cls.collection1,
                description="My description!",
                tags=["first-tag"],
            ),
            Entity.objects.create(
                name="Test entity 1",
                slug="test-entity-1",
                contributor=cls.contributor,
                collection=cls.collection2,
                tags=["first-tag", "second-tag"],
                description="My favourite test entity",
            ),
            Entity.objects.create(
                name="User test entity",
                slug="user-entity",
                contributor=cls.user,
                description="User's description",
                tags=["second-tag"],
            ),
            Entity.objects.create(
                name="Public test entity",
                slug="public-entity",
                contributor=cls.user,
            ),
        ]

        cls.rel_type = RelationType.objects.create(name="series", ordered=True)
        cls.relation1 = Relation.objects.create(
            type=cls.rel_type, collection=cls.collection1, contributor=cls.contributor
        )
        cls.relation1.entities.set(cls.entities[:1])
        cls.relation1.save()
        cls.relation2 = Relation.objects.create(
            type=cls.rel_type, collection=cls.collection2, contributor=cls.contributor
        )
        cls.relation2.entities.set(cls.entities[1:])
        cls.relation2.save()

        tzone = get_current_timezone()
        cls.entities[0].created = datetime.datetime(2016, 7, 30, 0, 59, tzinfo=tzone)
        cls.entities[0].save()
        cls.entities[1].created = datetime.datetime(2016, 7, 30, 1, 59, tzinfo=tzone)
        cls.entities[1].save()
        cls.entities[2].created = datetime.datetime(2016, 7, 30, 2, 59, tzinfo=tzone)
        cls.entities[2].save()

        cls.collection1.set_permission(Permission.OWNER, cls.admin)
        cls.collection1.set_permission(Permission.VIEW, cls.user)
        cls.collection2.set_permission(Permission.OWNER, cls.admin)
        cls.collection2.set_permission(Permission.VIEW, cls.user)
        cls.entities[2].set_permission(Permission.OWNER, cls.user)
        cls.entities[2].set_permission(Permission.VIEW, cls.group)
        cls.entities[3].set_permission(Permission.VIEW, get_anonymous_user())

        cls.viewset_class = EntityViewSet

    def test_filter_id(self):
        self._check_filter({"id": self.entities[0].pk}, [self.entities[0]])
        self._check_filter({"id": "420"}, [])
        self._check_filter(
            {"id__in": "{},{}".format(self.entities[0].pk, self.entities[1].pk)},
            self.entities[:2],
        )
        self._check_filter({"id__gt": str(self.entities[0].pk)}, self.entities[1:])

    def test_filter_slug(self):
        self._check_filter({"slug": "test-entity-0"}, [self.entities[0]])
        self._check_filter(
            {"slug__in": "test-entity-1,user-entity"}, self.entities[1:3]
        )

    def test_filter_name(self):
        self._check_filter({"name": "Test entity 1"}, [self.entities[1]])
        self._check_filter({"name": "Test entity"}, [])

        self._check_filter({"name__iexact": "test entity 1"}, [self.entities[1]])

        self._check_filter({"name__startswith": "Test entity"}, self.entities[:2])
        self._check_filter({"name__istartswith": "test entity"}, self.entities[:2])

        self._check_filter({"name__contains": "test"}, self.entities[2:])
        self._check_filter({"name__icontains": "test"}, self.entities)

    def test_filter_description(self):
        self._check_filter({"description__icontains": "Favourite"}, [self.entities[1]])
        self._check_filter({"description__contains": "Favourite"}, [])
        self._check_filter({"description__contains": "420"}, [])

    def test_filter_tags(self):
        self._check_filter({"tags": "first-tag"}, self.entities[:2])
        self._check_filter({"tags": "first-tag,second-tag"}, [self.entities[1]])

    def test_filter_contributor(self):
        self._check_filter({"contributor": str(self.contributor.pk)}, self.entities[:2])
        self._check_filter(
            {"contributor__in": "{},{}".format(self.contributor.pk, self.user.pk)},
            self.entities,
        )

    def test_filter_owners(self):
        self._check_filter({"owners": str(self.admin.pk)}, self.entities[:2])
        self._check_filter({"owners": "999999"}, [])

    def test_filter_contributor_name(self):
        # Filter by first name.
        self._check_filter({"contributor_name": "Joe"}, self.entities[:2])
        # Filter by last name.
        self._check_filter({"contributor_name": "Miller"}, self.entities[:2])
        # Filter by username.
        self._check_filter({"contributor_name": "contributor"}, self.entities[:2])
        # Filter by combination of first and last name.
        self._check_filter({"contributor_name": "John Williams"}, self.entities[2:])

    def test_filter_owners_name(self):
        # Filter by first name.
        self._check_filter({"owners_name": "James"}, self.entities[:2])
        # Filter by last name.
        self._check_filter({"owners_name": "Smith"}, self.entities[:2])
        # Filter by username.
        self._check_filter({"owners_name": "admin"}, self.entities[:2])
        # Filter by combination of first and last name.
        self._check_filter({"owners_name": "John Williams"}, [self.entities[2]])

    def test_filter_permission(self):
        self._check_filter({"permission": "edit"}, [self.entities[2]], user=self.user)
        self._check_filter(
            {"permission": "view"},
            self.entities,
            user=self.user,
        )
        self._check_filter({"permission": "owner"}, [self.entities[2]], user=self.user)

    def test_filter_created(self):
        self._check_filter({"created": "2016-07-30T00:59:00"}, self.entities[:1])
        self._check_filter({"created__gt": "2016-07-30T01:59:00"}, self.entities[2:])
        self._check_filter({"created__gte": "2016-07-30T01:59:00"}, self.entities[1:])
        self._check_filter({"created__lt": "2016-07-30T01:59:00"}, self.entities[:1])
        self._check_filter({"created__lte": "2016-07-30T01:59:00"}, self.entities[:2])

    def test_filter_modified(self):
        self._check_filter(
            {"modified": self.entities[0].modified.isoformat()}, self.entities[:1]
        )
        self._check_filter(
            {"modified__gt": self.entities[1].modified.isoformat()}, [self.entities[2]]
        )
        self._check_filter(
            {"modified__gte": self.entities[1].modified.isoformat()}, self.entities[1:3]
        )
        self._check_filter(
            {"modified__lt": self.entities[1].modified.isoformat()},
            self.entities[:1] + [self.entities[3]],
        )
        self._check_filter(
            {"modified__lte": self.entities[1].modified.isoformat()},
            self.entities[:2] + [self.entities[3]],
        )

    def test_filter_collection(self):
        self._check_filter({"collection": str(self.collection1.pk)}, [self.entities[0]])
        self._check_filter(
            {
                "collection__in": "{},{}".format(
                    self.collection1.pk, self.collection2.pk
                )
            },
            self.entities[:2],
        )
        self._check_filter({"collection": "999999"}, [])
        self._check_filter(
            {"collection__in": "{},{}".format(self.collection1.pk, "999999")},
            self.entities[:1],
        )
        self._check_filter(
            {"collection__isnull": True},
            self.entities[2:],
        )

    def test_filter_collection_name(self):
        self._check_filter({"collection__name": "My collection"}, [self.entities[0]])
        self._check_filter(
            {"collection__name__icontains": "Collection"}, self.entities[:2]
        )

    def test_filter_collection_slug(self):
        self._check_filter({"collection__slug": "my-collection"}, [self.entities[0]])
        self._check_filter(
            {"collection__slug__in": "my-collection,other-collection"},
            self.entities[:2],
        )

    def test_filter_text(self):
        # By name.
        self._check_filter({"text": "Test entity 1"}, [self.entities[1]])
        self._check_filter({"text": "Test"}, self.entities)
        self._check_filter({"text": "test"}, self.entities)
        self._check_filter({"text": "user"}, self.entities[2:])

        # By contributor.
        self._check_filter({"text": "joe"}, self.entities[:2])
        self._check_filter({"text": "Miller"}, self.entities[:2])

        # By owner.
        self._check_filter({"text": "james"}, self.entities[:2])
        self._check_filter({"text": "Smith"}, self.entities[:2])

        # By description.
        self._check_filter(
            {"text": "description"}, [self.entities[0], self.entities[2]]
        )
        self._check_filter({"text": "my"}, self.entities[:2])
        self._check_filter({"text": "my description"}, [self.entities[0]])
        self._check_filter({"text": "user"}, self.entities[2:])

        # By mixed fields.
        self._check_filter({"text": "test joe"}, self.entities[:2])
        self._check_filter({"text": "joe my description"}, [self.entities[0]])

        # Check that changes are applied immediately.
        self.entities[0].name = "Awesome new name"
        self.entities[0].save()

        self._check_filter({"text": "awesome"}, [self.entities[0]])

    def test_filter_mixed(self):
        self._check_filter(
            {"name": "Test entity 0", "owners_name": "James Smith"}, [self.entities[0]]
        )
        self._check_filter(
            {"text": "entity 0", "owners_name": "James Smith"}, [self.entities[0]]
        )

    def test_order_by_relevance(self):
        result = self._check_filter({"text": "test entity"}, self.entities)
        self.assertEqual(result.data[0]["id"], self.entities[1].pk)

        # Make another entity more important.
        entity = self.entities[2]
        entity.description = "This is test entity. My test entity rocks."
        entity.save()

        result = self._check_filter({"text": "test entity"}, self.entities)
        self.assertEqual(result.data[0]["id"], self.entities[2].pk)

        # Check that ordering can be overriden.
        result = self._check_filter(
            {"text": "test entity", "ordering": "id"}, self.entities
        )
        self.assertEqual(result.data[0]["id"], self.entities[0].pk)

    def test_order_by_collection_name(self):
        """Check entities can be ordered by the collection name."""
        request = factory.get("/", {"ordering": "collection__name"}, format="json")
        force_authenticate(request, self.admin)
        response = self.viewset(request)
        received = [
            (item["collection"] or {}).get("name", None) for item in response.data
        ]
        expected = ["My collection", "Other collection", None, None]
        self.assertEqual(received, expected)

        request = factory.get("/", {"ordering": "-collection__name"}, format="json")
        force_authenticate(request, self.admin)
        response = self.viewset(request)
        received = [
            (item["collection"] or {}).get("name", None) for item in response.data
        ]
        expected.reverse()
        self.assertEqual(received, expected)

    def test_filter_by_relation(self):
        self._check_filter({"relation_id": self.relation1.id}, self.entities[:1])
        self._check_filter({"relation_id": self.relation2.id}, self.entities[1:])
        self._check_filter({"relation_id": 42}, [])

    def test_filter_group(self):
        self._check_filter({"group": self.group.pk}, [self.entities[2]])

    def test_filter_shared_with_me(self):
        self._check_filter({"shared_with_me": True}, self.entities[:3], user=self.user)
        self._check_filter(
            {"shared_with_me": False}, [self.entities[3]], user=self.user
        )


class DataViewSetFiltersTest(BaseViewSetFiltersTest):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        tzone = get_current_timezone()

        descriptor_schema = DescriptorSchema.objects.create(
            slug="test-schema",
            version="1.0.0",
            contributor=cls.contributor,
            schema=[
                {
                    "name": "company",
                    "group": [
                        {
                            "name": "name",
                            "type": "basic:string:",
                            "required": False,
                        },
                        {
                            "name": "departments",
                            "type": "list:basic:string:",
                            "required": False,
                        },
                    ],
                }
            ],
        )

        cls.collection1 = Collection.objects.create(
            name="My collection",
            slug="my-collection",
            contributor=cls.contributor,
        )

        cls.collection2 = Collection.objects.create(
            name="Other collection",
            slug="other-collection",
            contributor=cls.contributor,
        )

        cls.entity1 = Entity.objects.create(
            name="My entity",
            slug="my-entity",
            collection=cls.collection1,
            contributor=cls.contributor,
        )

        cls.entity2 = Entity.objects.create(
            name="Other entity",
            slug="other-entity",
            collection=cls.collection2,
            contributor=cls.contributor,
        )

        cls.proc1 = Process.objects.create(
            type="data:test:process1:",
            name="First process",
            slug="test-process-1",
            version="1.0.0",
            contributor=cls.contributor,
            input_schema=[
                {"name": "input_data", "type": "data:test:", "required": False}
            ],
        )

        cls.proc2 = Process.objects.create(
            type="data:test:process2:",
            name="Second process",
            slug="test-process-2",
            version="1.0.0",
            contributor=cls.contributor,
            input_schema=[
                {"name": "input_data", "type": "data:test:", "required": False}
            ],
        )

        cls.data = [
            Data.objects.create(
                name="Data 0",
                slug="dataslug-0",
                contributor=cls.contributor,
                collection=cls.collection1,
                entity=cls.entity1,
                process=cls.proc1,
                status=Data.STATUS_RESOLVING,
                started=datetime.datetime(2016, 7, 31, 0, 0, tzinfo=tzone),
                finished=datetime.datetime(2016, 7, 31, 0, 30, tzinfo=tzone),
                tags=["first-tag"],
                descriptor_schema=descriptor_schema,
                descriptor={
                    "company": {
                        "name": "Genialis",
                        "departments": ["engineering", "operations"],
                    }
                },
            ),
            Data.objects.create(
                name="Data 1",
                slug="dataslug-1",
                contributor=cls.contributor,
                collection=cls.collection2,
                entity=cls.entity2,
                process=cls.proc1,
                status=Data.STATUS_DONE,
                started=datetime.datetime(2016, 7, 31, 1, 0, tzinfo=tzone),
                finished=datetime.datetime(2016, 7, 31, 1, 30, tzinfo=tzone),
                tags=["first-tag", "second-tag", "data"],
            ),
            Data.objects.create(
                name="User data",
                slug="dataslug-2",
                contributor=cls.user,
                process=cls.proc2,
                status=Data.STATUS_DONE,
                started=datetime.datetime(2016, 7, 31, 2, 0, tzinfo=tzone),
                finished=datetime.datetime(2016, 7, 31, 2, 30, tzinfo=tzone),
                tags=["second-tag"],
            ),
            Data.objects.create(
                name="Public data",
                slug="public-data",
                contributor=cls.user,
                process=cls.proc2,
                status=Data.STATUS_DONE,
                started=datetime.datetime(2016, 7, 31, 2, 0, tzinfo=tzone),
                finished=datetime.datetime(2016, 7, 31, 2, 30, tzinfo=tzone),
            ),
        ]

        cls.rel_type = RelationType.objects.create(name="series", ordered=True)
        cls.relation1 = Relation.objects.create(
            type=cls.rel_type, collection=cls.collection1, contributor=cls.contributor
        )
        cls.relation1.entities.set([cls.entity1, cls.entity2])
        cls.relation1.save()
        cls.relation2 = Relation.objects.create(
            type=cls.rel_type, collection=cls.collection2, contributor=cls.contributor
        )
        cls.relation2.entities.set([cls.entity2])
        cls.relation2.save()

        cls.data[0].created = datetime.datetime(2016, 7, 30, 0, 59, tzinfo=tzone)
        cls.data[0].save()
        cls.data[1].created = datetime.datetime(2016, 7, 30, 1, 59, tzinfo=tzone)
        cls.data[1].save()
        cls.data[2].created = datetime.datetime(2016, 7, 30, 2, 59, tzinfo=tzone)
        cls.data[2].save()

        cls.collection1.set_permission(Permission.OWNER, cls.admin)
        cls.collection1.set_permission(Permission.VIEW, cls.user)
        cls.collection2.set_permission(Permission.OWNER, cls.admin)
        cls.collection2.set_permission(Permission.VIEW, cls.user)
        cls.data[2].set_permission(Permission.OWNER, cls.user)
        cls.data[2].set_permission(Permission.VIEW, cls.group)
        cls.data[3].set_permission(Permission.VIEW, get_anonymous_user())

        cls.viewset_class = DataViewSet

    def test_filter_id(self):
        self._check_filter({"id": str(self.data[0].pk)}, [self.data[0]])
        self._check_filter({"id": 420}, [])
        self._check_filter(
            {"id__in": "{},{}".format(self.data[0].pk, self.data[2].pk)},
            [self.data[0], self.data[2]],
        )

    def test_filter_slug(self):
        self._check_filter({"slug": "dataslug-1"}, [self.data[1]])
        self._check_filter(
            {"slug__in": "dataslug-1,dataslug-2"}, [self.data[1], self.data[2]]
        )

    def test_filter_name(self):
        self._check_filter({"name": "Data 1"}, [self.data[1]])
        self._check_filter({"name": "Data"}, [])

        self._check_filter({"name__iexact": "data 1"}, [self.data[1]])

        self._check_filter({"name__startswith": "Data"}, self.data[:2])
        self._check_filter({"name__istartswith": "data"}, self.data[:2])

        self._check_filter({"name__contains": "Data"}, self.data[:2])
        self._check_filter({"name__icontains": "Data"}, self.data)

    def test_filter_tags(self):
        self._check_filter({"tags": "first-tag"}, self.data[:2])
        self._check_filter({"tags": "first-tag,second-tag"}, [self.data[1]])

    def test_filter_contributor(self):
        self._check_filter({"contributor": str(self.contributor.pk)}, self.data[:2])
        self._check_filter(
            {"contributor__in": "{},{}".format(self.contributor.pk, self.user.pk)},
            self.data,
        )

    def test_filter_owners(self):
        self._check_filter({"owners": str(self.admin.pk)}, self.data[:2])
        self._check_filter({"owners": "999999"}, [])

    def test_filter_contributor_name(self):
        # Filter by first name.
        self._check_filter({"contributor_name": "Joe"}, self.data[:2])
        # Filter by last name.
        self._check_filter({"contributor_name": "Miller"}, self.data[:2])
        # Filter by username.
        self._check_filter({"contributor_name": "contributor"}, self.data[:2])
        # Filter by combination of first and last name.
        self._check_filter({"contributor_name": "John Williams"}, self.data[2:])

    def test_filter_owners_name(self):
        # Filter by first name.
        self._check_filter({"owners_name": "James"}, self.data[:2])
        # Filter by last name.
        self._check_filter({"owners_name": "Smith"}, self.data[:2])
        # Filter by username.
        self._check_filter({"owners_name": "admin"}, self.data[:2])
        # Filter by combination of first and last name.
        self._check_filter({"owners_name": "John Williams"}, self.data[2:])

    def test_filter_permission(self):
        self._check_filter({"permission": "edit"}, self.data[2:], user=self.user)
        self._check_filter(
            {"permission": "view"},
            self.data,
            user=self.user,
        )
        self._check_filter({"permission": "owner"}, self.data[2:], user=self.user)

    def test_filter_created(self):
        self._check_filter({"created": "2016-07-30T00:59:00"}, [self.data[0]])
        self._check_filter({"created__gt": "2016-07-30T01:59:00"}, self.data[2:])
        self._check_filter({"created__gte": "2016-07-30T01:59:00"}, self.data[1:])
        self._check_filter({"created__lt": "2016-07-30T01:59:00"}, self.data[:1])
        self._check_filter({"created__lte": "2016-07-30T01:59:00"}, self.data[:2])

    def test_filter_modified(self):
        self._check_filter(
            {"modified": self.data[0].modified.isoformat()}, [self.data[0]]
        )
        self._check_filter(
            {"modified__gt": self.data[1].modified.isoformat()}, [self.data[2]]
        )
        self._check_filter(
            {"modified__gte": self.data[1].modified.isoformat()}, self.data[1:3]
        )
        self._check_filter(
            {"modified__lt": self.data[1].modified.isoformat()},
            self.data[:1] + [self.data[3]],
        )
        self._check_filter(
            {"modified__lte": self.data[1].modified.isoformat()},
            self.data[:2] + [self.data[3]],
        )

    def test_filter_started(self):
        self._check_filter({"started": "2016-07-31T00:00:00"}, [self.data[0]])
        self._check_filter({"started__gt": "2016-07-31T01:00:00"}, self.data[2:])
        self._check_filter({"started__gte": "2016-07-31T01:00:00"}, self.data[1:])
        self._check_filter({"started__lt": "2016-07-31T01:00:00"}, self.data[:1])
        self._check_filter({"started__lte": "2016-07-31T01:00:00"}, self.data[:2])

    def test_filter_finished(self):
        self._check_filter({"finished": "2016-07-31T00:30:00"}, [self.data[0]])
        self._check_filter({"finished__gt": "2016-07-31T01:30:00"}, self.data[2:])
        self._check_filter({"finished__gte": "2016-07-31T01:30:00"}, self.data[1:])
        self._check_filter({"finished__lt": "2016-07-31T01:30:00"}, self.data[:1])
        self._check_filter({"finished__lte": "2016-07-31T01:30:00"}, self.data[:2])

    def test_filter_collection(self):
        self._check_filter({"collection": str(self.collection1.pk)}, [self.data[0]])
        self._check_filter(
            {
                "collection__in": "{},{}".format(
                    self.collection1.pk, self.collection2.pk
                )
            },
            self.data[:2],
        )
        self._check_filter({"collection": "999999"}, [])
        self._check_filter(
            {"collection__in": "{},{}".format(self.collection1.pk, "999999")},
            self.data[:1],
        )
        self._check_filter(
            {"collection__isnull": True},
            self.data[2:],
        )

    def test_mixed_process_version(self):
        """Test processes with mixed versions are shown.

        When multiple data objects with different versions of the process with same
        slug are in the filter result both process versions must be returned.
        """
        # Create data object with never version of the test-process-1 process.
        process = Process.objects.create(
            slug="test-process-1", version="2.0.0", contributor=self.contributor
        )
        Data.objects.create(
            contributor=self.contributor, collection=self.collection1, process=process
        ),
        request = factory.get("/", {"collection": self.collection1.pk}, format="json")
        force_authenticate(request, self.admin)
        response = self.viewset(request)
        self.assertEqual(len(response.data), 2)
        self.assertCountEqual(
            ["1.0.0", "2.0.0"],
            [entry["process"]["version"] for entry in response.data],
        )

    def test_mixed_descriptor_schema_version(self):
        """Test processes with mixed descriptor schema versions are shown.

        When multiple data objects with different versions of the descriptor schema
        with same slug are in the filter result both process versions must be returned.
        """
        descriptor_schema = DescriptorSchema.objects.create(
            slug="test-schema",
            version="2.0.0",
            contributor=self.contributor,
            schema=[
                {
                    "name": "company",
                    "group": [
                        {
                            "name": "name",
                            "type": "basic:string:",
                            "required": False,
                        },
                        {
                            "name": "departments",
                            "type": "list:basic:string:",
                            "required": False,
                        },
                    ],
                }
            ],
        )
        Data.objects.create(
            contributor=self.contributor,
            collection=self.collection1,
            process=self.proc1,
            descriptor_schema=descriptor_schema,
        ),
        request = factory.get("/", {"collection": self.collection1.pk}, format="json")
        force_authenticate(request, self.admin)
        response = self.viewset(request)
        self.assertEqual(len(response.data), 2)
        self.assertCountEqual(
            ["1.0.0", "2.0.0"],
            [entry["descriptor_schema"]["version"] for entry in response.data],
        )

    def test_filter_collection_name(self):
        self._check_filter({"collection__name": "My collection"}, [self.data[0]])
        self._check_filter({"collection__name__icontains": "Collection"}, self.data[:2])

    def test_filter_collection_slug(self):
        self._check_filter({"collection__slug": "my-collection"}, [self.data[0]])
        self._check_filter(
            {"collection__slug__in": "my-collection,other-collection"}, self.data[:2]
        )

    def test_filter_entity(self):
        self._check_filter({"entity": str(self.entity1.pk)}, [self.data[0]])
        self._check_filter(
            {"entity__in": "{},{}".format(self.entity1.pk, self.entity2.pk)},
            self.data[:2],
        )
        self._check_filter({"entity": "999999"}, [])
        self._check_filter(
            {"entity__in": "{},{}".format(self.entity1.pk, "999999")},
            [self.data[0]],
        )
        self._check_filter(
            {"entity__in": "{},{}".format(self.entity1.pk, self.entity2.pk)},
            self.data[:2],
        )
        self._check_filter(
            {"entity__isnull": True},
            self.data[2:],
        )

    def test_filter_entity_name(self):
        self._check_filter({"entity__name": "My entity"}, [self.data[0]])
        self._check_filter({"entity__name__icontains": "Entity"}, self.data[:2])

    def test_filter_entity_slug(self):
        self._check_filter({"entity__slug": "my-entity"}, [self.data[0]])
        self._check_filter(
            {"entity__slug__in": "my-entity,other-entity"}, self.data[:2]
        )

    def test_filter_type(self):
        self._check_filter({"type": "data:"}, self.data)
        self._check_filter({"type": "data:test:"}, self.data)
        self._check_filter({"type": "data:test:process1"}, self.data[:2])
        self._check_filter({"type": "data:test:process1:"}, self.data[:2])
        self._check_filter({"type__exact": "data:test"}, [])
        self._check_filter({"type__exact": "data:test:process1:"}, self.data[:2])

    def test_filter_status(self):
        self._check_filter({"status": "OK"}, self.data[1:])
        self._check_filter({"status": "RE"}, [self.data[0]])
        self._check_filter({"status__in": "OK,RE"}, self.data)

    def test_filter_process(self):
        self._check_filter({"process": str(self.proc1.pk)}, self.data[:2])
        self._check_filter({"process": str(self.proc2.pk)}, self.data[2:])
        self._check_filter({"process": "999999"}, [])

    def test_filter_process_name(self):
        self._check_filter({"process__name": "First process"}, self.data[:2])
        self._check_filter({"process__name__iexact": "first process"}, self.data[:2])
        self._check_filter({"process__name__startswith": "First"}, self.data[:2])
        self._check_filter({"process__name": "first process"}, [])

    def test_filter_process_slug(self):
        self._check_filter({"process__slug": "test-process-1"}, self.data[:2])
        self._check_filter({"process__slug": "test-process-2"}, self.data[2:])
        self._check_filter({"process__slug": "test-process"}, [])

    def test_filter_text(self):
        # By name.
        self._check_filter({"text": "Data 1"}, [self.data[1]])
        self._check_filter({"text": "Data"}, self.data)
        self._check_filter({"text": "data"}, self.data)

        # By contributor.
        self._check_filter({"text": "joe"}, self.data[:2])
        self._check_filter({"text": "Miller"}, self.data[:2])

        # By owner.
        self._check_filter({"text": "james"}, self.data[:2])
        self._check_filter({"text": "Smith"}, self.data[:2])

        # By process name.
        self._check_filter({"text": "first"}, self.data[:2])
        self._check_filter({"text": "process"}, self.data)

        # By descriptor.
        self._check_filter({"text": "genialis"}, [self.data[0]])
        self._check_filter({"text": "engineering"}, [self.data[0]])

        # By mixed fields.
        self._check_filter({"text": "data joe"}, self.data[:2])
        self._check_filter({"text": "joe first"}, self.data[:2])

        # Check that changes are applied immediately.
        self.data[0].name = "Awesome new name"
        self.data[0].save()

        self._check_filter({"text": "awesome"}, [self.data[0]])

    def test_filter_mixed(self):
        self._check_filter(
            {"name": "Data 0", "owners_name": "James Smith"}, [self.data[0]]
        )
        self._check_filter(
            {"text": "Data 0", "owners_name": "James Smith"}, [self.data[0]]
        )

    def test_order_by_relevance(self):
        result = self._check_filter({"text": "data"}, self.data)
        self.assertEqual(result.data[0]["id"], self.data[1].pk)

        # Make another entity more important.
        data = self.data[2]
        data.name = "Data data data"
        data.save()

        result = self._check_filter({"text": "data"}, self.data)
        self.assertEqual(result.data[0]["id"], self.data[2].pk)

        # Check that ordering can be overriden.
        result = self._check_filter({"text": "data", "ordering": "id"}, self.data)
        self.assertEqual(result.data[0]["id"], self.data[0].pk)

    def test_ordering_entity_name(self):
        """Check that data can be ordered by the entity name."""
        expected = ["My entity", "Other entity", None, None]

        request = factory.get("/", {"ordering": "entity__name"}, format="json")
        force_authenticate(request, self.admin)
        response = self.viewset(request)
        received = [(item.get("entity") or {}).get("name") for item in response.data]
        self.assertTrue(status.is_success(response.status_code))
        self.assertEqual(expected, received)

        request = factory.get("/", {"ordering": "-entity__name"}, format="json")
        force_authenticate(request, self.admin)
        response = self.viewset(request)
        received = [(item.get("entity") or {}).get("name") for item in response.data]
        self.assertTrue(status.is_success(response.status_code))
        self.assertEqual(list(reversed(expected)), received)

    def test_nonexisting_parameter(self):
        response = self._check_filter({"foo": "bar"}, [], expected_status_code=400)

        self.assertRegex(
            str(response.data["__all__"]),
            r"Unsupported parameter\(s\): foo. Please use a combination of: .*",
        )

    def test_filter_by_relation(self):
        self._check_filter(
            {"relation_id": self.relation1.id}, [self.data[0], self.data[1]]
        )
        self._check_filter({"relation_id": self.relation2.id}, [self.data[1]])
        self._check_filter({"relation_id": 42}, [])

    def test_filter_group(self):
        self._check_filter({"group": self.group.pk}, [self.data[2]])

    def test_filter_shared_with_me(self):
        self._check_filter({"shared_with_me": True}, self.data[:3], user=self.user)
        self._check_filter({"shared_with_me": False}, [self.data[3]], user=self.user)


class DescriptorSchemaViewSetFiltersTest(BaseViewSetFiltersTest):
    def setUp(self):
        self.viewset_class = DescriptorSchemaViewSet
        super().setUp()

        tzone = get_current_timezone()

        self.ds1 = DescriptorSchema.objects.create(
            contributor=self.user, slug="slug1", name="ds1"
        )
        self.ds1.created = datetime.datetime(2015, 7, 28, 11, 57, tzinfo=tzone)
        self.ds1.save()

        self.ds2 = DescriptorSchema.objects.create(
            contributor=self.user, slug="slug2", name="ds2"
        )
        self.ds2.created = datetime.datetime(2016, 8, 29, 12, 58, 0, tzinfo=tzone)
        self.ds2.save()

        self.ds3 = DescriptorSchema.objects.create(
            contributor=self.admin, slug="slug3", name="ds3"
        )
        self.ds3.created = datetime.datetime(2017, 9, 30, 13, 59, 0, tzinfo=tzone)
        self.ds3.save()

    def test_id(self):
        self._check_filter({"id": self.ds1.pk}, [self.ds1])
        self._check_filter({"id": self.ds2.pk}, [self.ds2])
        self._check_filter({"id__in": self.ds2.pk}, [self.ds2])
        self._check_filter(
            {"id__in": "{},{}".format(self.ds1.pk, self.ds2.pk)}, [self.ds1, self.ds2]
        )
        self._check_filter({"id__gt": self.ds1.pk}, [self.ds2, self.ds3])

    def test_slug(self):
        self._check_filter({"slug": "slug1"}, [self.ds1])
        self._check_filter({"slug": "slug2"}, [self.ds2])
        self._check_filter({"slug__in": "slug"}, [])
        self._check_filter({"slug__in": "slug2,slug"}, [self.ds2])

    def test_name(self):
        self._check_filter({"name": "ds1"}, [self.ds1])
        self._check_filter({"name": "ds2"}, [self.ds2])
        self._check_filter({"name__startswith": "ds"}, [self.ds1, self.ds2, self.ds3])
        self._check_filter({"name__startswith": "ds2"}, [self.ds2])
        self._check_filter({"name__startswith": "dsx"}, [])
        self._check_filter({"name__icontains": "2"}, [self.ds2])
        self._check_filter({"name__icontains": "DS"}, [self.ds1, self.ds2, self.ds3])

    def test_contributor(self):
        self._check_filter({"contributor": self.user.pk}, [self.ds1, self.ds2])
        self._check_filter({"contributor": self.admin.pk}, [self.ds3])
        self._check_filter({"contributor": "999999"}, [])
        self._check_filter({"contributor__in": self.admin.pk}, [self.ds3])
        self._check_filter(
            {"contributor__in": "999999,{}".format(self.admin.pk)}, [self.ds3]
        )
        self._check_filter(
            {"contributor__in": "{},{}".format(self.user.pk, self.admin.pk)},
            [self.ds1, self.ds2, self.ds3],
        )

    def test_created(self):
        self._check_filter({"created": "2015-07-28T11:57:00"}, [self.ds1])
        self._check_filter({"created": "2016-08-29T12:58:00.000000"}, [self.ds2])
        self._check_filter({"created": "2017-09-30T13:59:00.000000Z"}, [self.ds3])
        self._check_filter({"created": "2017-09-30T13:59:00.000000+0000"}, [self.ds3])
        self._check_filter({"created": "2017-09-30T13:59:00.000000+0000"}, [self.ds3])

        self._check_filter({"created__date": "2016-08-29"}, [self.ds2])
        self._check_filter({"created__time": "12:58:00"}, [self.ds2])

    def test_modified(self):
        now = self.ds1.modified
        self._check_filter(
            {"modified": now.strftime("%Y-%m-%dT%H:%M:%S.%f%z")}, [self.ds1]
        )

    def test_nonexisting_parameter(self):
        response = self._check_filter(
            {"foo": "bar"}, [self.ds1], expected_status_code=400
        )

        self.assertRegex(
            str(response.data["__all__"][0]),
            r"Unsupported parameter\(s\): foo. Please use a combination of: .*",
        )


class ProcessViewSetFiltersTest(BaseViewSetFiltersTest):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()

        cls.proc_1 = Process.objects.create(
            contributor=cls.contributor,
            type="data:alignment:bam:",
            category="analyses:alignment:",
            scheduling_class="BA",
        )

        cls.proc_2 = Process.objects.create(
            contributor=cls.contributor,
            type="data:expression:",
            category="analyses:",
            scheduling_class="IN",
        )

        cls.viewset_class = ProcessViewSet

    def test_category(self):
        self._check_filter({"category": "analyses:"}, [self.proc_1, self.proc_2])
        self._check_filter({"category": "analyses:alignment"}, [self.proc_1])

    def test_type(self):
        self._check_filter({"type": "data:"}, [self.proc_1, self.proc_2])
        self._check_filter({"type": "data:alignment:bam"}, [self.proc_1])
        self._check_filter({"type": "data:expression"}, [self.proc_2])

    def test_scheduling_class(self):
        self._check_filter({"scheduling_class": "BA"}, [self.proc_1])
        self._check_filter({"scheduling_class": "IN"}, [self.proc_2])

    def test_is_active(self):
        self._check_filter({"is_active": "true"}, [self.proc_1, self.proc_2])
        self._check_filter({"is_active": "false"}, [])

        self.proc_1.is_active = False
        self.proc_1.save()

        self._check_filter({"is_active": "true"}, [self.proc_2])
        self._check_filter({"is_active": "false"}, [self.proc_1])


class RelationViewSetFiltersTest(BaseViewSetFiltersTest):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()

        cls.collection_1 = Collection.objects.create(
            contributor=cls.contributor,
            tags=["first-tag"],
        )

        cls.collection_2 = Collection.objects.create(
            contributor=cls.contributor,
            tags=["second-tag"],
        )

        cls.rel_type = RelationType.objects.create(name="series", ordered=True)
        cls.relation_1 = Relation.objects.create(
            type=cls.rel_type, collection=cls.collection_1, contributor=cls.contributor
        )

        cls.relation_2 = Relation.objects.create(
            type=cls.rel_type, collection=cls.collection_2, contributor=cls.contributor
        )

        cls.viewset_class = RelationViewSet

    def test_filter_tags(self):
        self._check_filter({"tags": "first-tag"}, [self.relation_1])



""".. Ignore pydocstyle D400.

==========
Dispatcher
==========

"""

import asyncio
import json
import logging
import os
import shlex
import shutil
from contextlib import suppress
from importlib import import_module
from pathlib import Path
from typing import List, Optional, Tuple, Union

from channels.db import database_sync_to_async
from channels.exceptions import ChannelFull
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured, PermissionDenied
from django.db import IntegrityError, connection, models, transaction
from django.db.models import OuterRef, Q, Subquery
from django.utils.timezone import now
from zmq import curve_keypair

from resolwe.flow.engine import InvalidEngineError, load_engines
from resolwe.flow.execution_engines import ExecutionError
from resolwe.flow.executors.constants import PROCESSING_VOLUME_NAME
from resolwe.flow.models import Data, DataDependency, Process, Worker
from resolwe.flow.models.utils import referenced_files
from resolwe.storage import settings as storage_settings
from resolwe.storage.connectors import DEFAULT_CONNECTOR_PRIORITY, connectors
from resolwe.storage.models import (
    AccessLog,
    FileStorage,
    ReferencedPath,
    StorageLocation,
)
from resolwe.utils import BraceMessage as __

from . import consumer, state
from .protocol import WorkerProtocol

logger = logging.getLogger(__name__)

DEFAULT_CONNECTOR = "resolwe.flow.managers.workload_connectors.local"


class SettingsJSONifier(json.JSONEncoder):
    """Customized JSON encoder, coercing all unknown types into strings.

    Needed due to the class hierarchy coming out of the database,
    which can't be serialized using the vanilla json encoder.
    """

    def default(self, o):
        """Try default; otherwise, coerce the object into a string."""
        try:
            return super().default(o)
        except TypeError:
            return str(o)


class Manager:
    """The manager handles process job dispatching.

    Each :class:`~resolwe.flow.models.Data` object that's still waiting
    to be resolved is dispatched to a concrete workload management
    system (such as Celery or SLURM). The specific manager for that
    system (descended from
    :class:`~resolwe.flow.managers.workload_connectors.base.BaseConnector`)
    then handles actual job setup and submission. The job itself is an
    executor invocation; the executor then in turn sets up a safe and
    well-defined environment within the workload manager's task in which
    the process is finally run.
    """

    def discover_engines(self):
        """Discover configured engines.

        :param executor: Optional executor module override
        """
        executor = getattr(settings, "FLOW_EXECUTOR", {}).get(
            "NAME", "resolwe.flow.executors.local"
        )
        self.executor = self.load_executor(executor)
        logger.info(
            __(
                "Loaded '{}' executor.",
                str(self.executor.__class__.__module__).replace(".prepare", ""),
            )
        )

        expression_engines = getattr(
            settings,
            "FLOW_EXPRESSION_ENGINES",
            ["resolwe.flow.expression_engines.jinja"],
        )
        self.expression_engines = self.load_expression_engines(expression_engines)
        logger.info(
            __(
                "Found {} expression engines: {}",
                len(self.expression_engines),
                ", ".join(self.expression_engines.keys()),
            )
        )

        execution_engines = getattr(
            settings, "FLOW_EXECUTION_ENGINES", ["resolwe.flow.execution_engines.bash"]
        )
        self.execution_engines = self.load_execution_engines(execution_engines)
        logger.info(
            __(
                "Found {} execution engines: {}",
                len(self.execution_engines),
                ", ".join(self.execution_engines.keys()),
            )
        )

    async def drain_messages(self):
        """Drain Django Channel messages."""
        await consumer.run_consumer(timeout=1)

    def __init__(self, *args, **kwargs):
        """Initialize arguments."""
        self.discover_engines()
        # Value is the indicator whether we are running in sync mode.
        self._sync_finished_event = None
        # The couter for number of messages that are being processed.
        # Used with sync mode to know when to remove the runtime
        # barrier.
        self._messages_processing = 0

        # Ensure there is only one manager instance per process. This
        # is required as other parts of the code access the global
        # manager instance.
        with suppress(ImportError):
            from resolwe.flow import managers

            assert not hasattr(managers, "manager")

        self.scheduling_class_map = dict(Process.SCHEDULING_CLASS_CHOICES)

        # Check settings for consistency.
        flow_manager = getattr(settings, "FLOW_MANAGER", {})
        if "DISPATCHER_MAPPING" in flow_manager and "NAME" in flow_manager:
            raise ImproperlyConfigured(
                "Key 'DISPATCHER_MAPPING' conflicts with key 'NAME' in FLOW_MANAGER settings."
            )

        if "DISPATCHER_MAPPING" in flow_manager:
            mapping = flow_manager["DISPATCHER_MAPPING"]

            scheduling_classes = set(self.scheduling_class_map.values())
            map_keys = set(mapping.keys())
            class_difference = scheduling_classes - map_keys
            if class_difference:
                raise ImproperlyConfigured(
                    "Dispatcher manager mapping settings incomplete, missing {}.".format(
                        class_difference
                    )
                )
            connector_list = [mapping[klass] for klass in scheduling_classes]
        else:
            connector_list = [flow_manager.get("NAME", DEFAULT_CONNECTOR)]

        # Store the whitelist and blacklist for later use.
        self._processes_allow = getattr(settings, "FLOW_PROCESSES_ALLOW_LIST", None)
        self._processes_ignore = getattr(settings, "FLOW_PROCESSES_IGNORE_LIST", None)

        # Pre-load all needed connectors.
        self.connectors = {}
        for module_name in connector_list:
            connector_module = import_module(module_name)
            self.connectors[module_name] = connector_module.Connector()

        logger.info(
            __(
                "Found {} workload connectors: {}",
                len(self.connectors),
                ", ".join(self.connectors.keys()),
            )
        )

        super().__init__(*args, **kwargs)

    def _include_environment_variables(self, program: str, executor_vars: dict) -> str:
        """Include environment variables in program."""
        env_vars = {
            "RESOLWE_HOST_URL": getattr(settings, "RESOLWE_HOST_URL", "localhost"),
        }
        set_env = getattr(settings, "FLOW_EXECUTOR", {}).get("SET_ENV", {})
        env_vars.update(executor_vars)
        env_vars.update(set_env)

        export_commands = [
            "export {}={}".format(key, shlex.quote(value))
            for key, value in env_vars.items()
        ]
        return os.linesep.join(export_commands) + os.linesep + program

    def run(self, data: Data, argv: List):
        """Select a concrete connector and run the process through it.

        :param data: The :class:`~resolwe.flow.models.Data` object that
            is to be run.
        :param argv: The argument vector used to spawn the executor.
        """
        process_scheduling = self.scheduling_class_map[data.process.scheduling_class]
        if "DISPATCHER_MAPPING" in getattr(settings, "FLOW_MANAGER", {}):
            class_name = settings.FLOW_MANAGER["DISPATCHER_MAPPING"][process_scheduling]
        else:
            class_name = getattr(settings, "FLOW_MANAGER", {}).get(
                "NAME", DEFAULT_CONNECTOR
            )

        data.scheduled = now()
        data.save(update_fields=["scheduled"])

        workload_class = class_name.rsplit(".", maxsplit=1)[1]
        host, port, protocol = self._get_listener_settings(data, workload_class)
        argv[-1] += " {} {} {}".format(host, port, protocol)

        return self.connectors[class_name].submit(data, argv)

    def _get_data_connector_name(self) -> str:
        """Return storage connector that will be used for new data object.

        The current implementation returns the connector with the lowest
        priority.
        """
        return min(
            (connector.priority, connector.name) for connector in connectors.values()
        )[1]

    def _prepare_data_dir(self, data: Data):
        """Prepare destination directory where the data will live.

        :param data: The :class:`~resolwe.flow.models.Data` object for
            which to prepare the private execution directory.
        :return: The prepared data directory path.
        :rtype: str
        """
        logger.debug(__("Preparing data directory for Data with id {}.", data.id))
        connector_name = self._get_data_connector_name()
        with transaction.atomic():
            # Create Worker object and set its status to preparing if needed.
            if not Worker.objects.filter(data=data).exists():
                public_key, private_key = curve_keypair()
                Worker.objects.get_or_create(
                    data=data,
                    status=Worker.STATUS_PREPARING,
                    public_key=public_key,
                    private_key=private_key,
                )

            file_storage = FileStorage.objects.create()
            # Data produced by the processing container will be uploaded to the
            # created location.
            data_location = StorageLocation.objects.create(
                file_storage=file_storage,
                url=str(file_storage.id),
                status=StorageLocation.STATUS_PREPARING,
                connector_name=connector_name,
            )
            file_storage.data.add(data)

            # Reference 'special' files.
            for file_ in referenced_files(data, include_descriptor=False):
                referenced_path = ReferencedPath.objects.create(path=file_)
                referenced_path.storage_locations.add(data_location)

        dir_mode = getattr(settings, "FLOW_EXECUTOR", {}).get("DATA_DIR_MODE", 0o755)
        connectors[connector_name].prepare_url(data_location.url, dir_mode=dir_mode)

    def _get_listener_settings(
        self, data: Data, workload_class_name: str
    ) -> Tuple[str, int, str]:
        """Return the listener address, port and protocol."""
        listener_settings = getattr(settings, "FLOW_EXECUTOR", {}).get(
            "LISTENER_CONNECTION", {}
        )
        container_listener_connection = getattr(
            settings,
            "COMMUNICATION_CONTAINER_LISTENER_CONNECTION",
            {"local": "172.17.0.1"},
        )

        # Return address associated with the key workload_class_name or first
        # one if the key is not present.
        if workload_class_name in container_listener_connection:
            host = container_listener_connection[workload_class_name]
        else:
            host = next(iter(container_listener_connection.values()))
        return (
            host,
            listener_settings.get("port", 53893),
            listener_settings.get("protocol", "tcp"),
        )

    async def handle_control_event(self, message: dict):
        """Handle the control event.

        The method is called from the channels layer when there is nome change
        either in the state of the Data object of the executors have finished
        with processing.

        When running in sync state check that all database objects are in
        final state before raising the execution_barrier.

        Channels layer callback, do not call directly.
        """

        def workers_finished():
            unfinished_workers = Worker.objects.exclude(
                status__in=Worker.FINAL_STATUSES
            )
            return not (unfinished_workers.exists())

        cmd = message[WorkerProtocol.COMMAND]
        logger.debug(__("Manager worker got channel command '{}'.", message))

        try:
            if cmd == WorkerProtocol.COMMUNICATE:
                await database_sync_to_async(self._data_scan, thread_sensitive=False)(
                    **message[WorkerProtocol.COMMUNICATE_EXTRA]
                )

            def purge_secrets_and_local_data(data_id: int) -> Data:
                """Purge secrets and return the Data object.

                :raises Data.DoesNotExist: when Data object does not exist.
                :raises FileStorage.DoesNotExist: when FileStorage object does not
                    exist.

                """
                data = Data.objects.get(pk=data_id)
                with suppress(Worker.DoesNotExist):
                    worker = Worker.objects.get(data=data)
                    worker.status = Worker.STATUS_COMPLETED
                    worker.save()

                subpath = FileStorage.objects.get(data__id=data_id).subpath
                volume = storage_settings.FLOW_VOLUMES[PROCESSING_VOLUME_NAME]
                if volume["type"] == "host_path":
                    processing_dir = Path(volume["config"]["path"]) / subpath
                    if processing_dir.is_dir():
                        shutil.rmtree(processing_dir)

                return data

            if cmd in [WorkerProtocol.FINISH, WorkerProtocol.ABORT]:
                self._messages_processing += 1
                data_id = message.get(WorkerProtocol.DATA_ID)
                if data_id is not None:
                    with suppress(
                        Data.DoesNotExist,
                        Process.DoesNotExist,
                        FileStorage.DoesNotExist,
                        AttributeError,
                    ):
                        await database_sync_to_async(
                            purge_secrets_and_local_data, thread_sensitive=False
                        )(data_id)
        except Exception:
            logger.exception(
                "Unknown error occured while processing communicate control command."
            )
            raise
        finally:
            self._messages_processing -= 1
            # No additional messages are being procesed.
            # Check if workers are finished with the processing.
            logger.debug(__("Dispatcher checking final condition"))
            logger.debug(__("Message counter: {}", self._messages_processing))

            if self._sync_finished_event is not None and self._messages_processing == 0:
                if await database_sync_to_async(
                    workers_finished, thread_sensitive=False
                )():
                    self._sync_finished_event.set()

    async def execution_barrier(self):
        """Wait for executors to finish.

        At least one must finish after this point to avoid a deadlock.
        """
        consumer_future = asyncio.ensure_future(consumer.run_consumer())
        assert self._sync_finished_event is not None
        await self._sync_finished_event.wait()
        self._sync_finished_event = None
        await consumer.exit_consumer()
        await consumer_future

    async def communicate(self, data_id=None, run_sync=False):
        """Scan database for resolving Data objects and process them.

        This is submitted as a task to the manager's channel workers.

        :param data_id: Optional id of Data object which (+ its
            children) should be processes. If it is not given, all
            resolving objects are processed.
        :param run_sync: If ``True``, wait until all processes spawned
            from this point on have finished processing.
        """
        first_sync_call = False
        if run_sync and self._sync_finished_event is None:
            first_sync_call = True
            self._sync_finished_event = asyncio.Event()

        self._messages_processing += 1
        logger.debug(
            __(
                "Manager sending communicate command on '{}' triggered by Data with id {}.",
                state.MANAGER_CONTROL_CHANNEL,
                data_id,
            )
        )
        try:
            await consumer.send_event(
                {
                    WorkerProtocol.COMMAND: WorkerProtocol.COMMUNICATE,
                    WorkerProtocol.COMMUNICATE_EXTRA: {
                        "data_id": data_id,
                    },
                }
            )
        except ChannelFull:
            logger.exception(
                "ChannelFull error occurred while sending communicate message."
            )
        except:
            logger.exception(
                "Unknown error occurred while sending communicate message."
            )

        if first_sync_call:
            logger.debug(
                __(
                    "Manager on channel '{}' entering synchronization block.",
                    state.MANAGER_CONTROL_CHANNEL,
                )
            )
            await self.execution_barrier()
            logger.debug(
                __(
                    "Manager on channel '{}' exiting synchronization block.",
                    state.MANAGER_CONTROL_CHANNEL,
                )
            )

    def _lock_inputs_local_storage_locations(self, data: Data):
        """Lock storage locations for inputs.

        Lock storage locations of inputs so they are not deleted while data
        object is processing.
        """
        data_connectors = connectors.for_storage("data")
        mountable_data_connectors = [
            connector for connector in data_connectors if connector.mountable
        ]
        priority_range = data_connectors[-1].priority - data_connectors[0].priority + 1

        connector_priorities = {
            connector.name: connector.priority for connector in data_connectors
        }

        # Prefer mountable locations but keep their relations intact.
        for connector in mountable_data_connectors:
            connector_priorities[connector.name] -= priority_range

        whens = [
            models.When(connector_name=connector_name, then=priority)
            for connector_name, priority in connector_priorities.items()
        ]

        storage_location_subquery = (
            StorageLocation.objects.filter(file_storage_id=OuterRef("file_storage_id"))
            .annotate(
                priority=models.Case(
                    *whens,
                    default=DEFAULT_CONNECTOR_PRIORITY,
                    output_field=models.IntegerField(),
                )
            )
            .order_by("priority")
            .values_list("id", flat=True)[:1]
        )

        file_storages = (
            DataDependency.objects.filter(child=data, kind=DataDependency.KIND_IO)
            .values_list("parent__location", flat=True)
            .distinct()
        )

        storage_locations = (
            StorageLocation.objects.filter(file_storage__in=file_storages)
            .filter(pk__in=Subquery(storage_location_subquery))
            .values_list("id", flat=True)
        )

        AccessLog.objects.bulk_create(
            [
                AccessLog(
                    storage_location_id=storage_location,
                    reason="Input for data with id {}".format(data.id),
                    cause=data,
                )
                for storage_location in storage_locations
            ]
        )

    def _data_execute(self, data: Data):
        """Execute the Data object.

        The activities carried out here include target directory
        preparation, executor copying, setting serialization and actual
        execution of the object.

        :param data: The :class:`~resolwe.flow.models.Data` object to
            execute.
        """
        logger.debug(__("Manager preparing Data with id {} for processing.", data.id))

        # Prepare the executor's environment.
        try:
            self._prepare_data_dir(data)

            executor_module = ".{}".format(
                getattr(settings, "FLOW_EXECUTOR", {})
                .get("NAME", "resolwe.flow.executors.local")
                .rpartition(".executors.")[-1]
            )
            self._lock_inputs_local_storage_locations(data)

            argv = [
                "/bin/sh",
                "-c",
                getattr(settings, "FLOW_EXECUTOR", {}).get(
                    "PYTHON", "/usr/bin/env python"
                )
                + " -m executors "
                + executor_module
                + " {}".format(data.pk),
            ]
            self.executor.prepare_for_execution(data)
        except PermissionDenied as error:
            data.status = Data.STATUS_ERROR
            data.process_error.append("Permission denied for process: {}".format(error))
            data.save()
            if hasattr(data, "worker"):
                data.worker.status = Worker.STATUS_ERROR_PREPARING
                data.worker.save()
            return
        except OSError as err:
            logger.exception(
                __(
                    "OSError occurred while preparing data {} (will skip): {}",
                    data.id,
                    err,
                )
            )
            if hasattr(data, "worker"):
                data.worker.status = Worker.STATUS_ERROR_PREPARING
                data.worker.save()
            return

        # Hand off to the run() method for execution.
        logger.info(__("Running executor for data with id {}", data.pk))
        self.run(data, argv)

    def _data_scan(self, data_id: Optional[int] = None, **kwargs):
        """Scan for new Data objects and execute them.

        :param data_id: Optional id of Data object which (+ its
            children) should be scanned. If it is not given, all
            resolving objects are processed.
        :param executor: The fully qualified name of the executor to use
            for all :class:`~resolwe.flow.models.Data` objects
            discovered in this pass.
        """

        def process_data_object(data: Data):
            """Process a single data object."""
            # Lock for update. Note that we want this transaction to be as short as possible in
            # order to reduce contention and avoid deadlocks. This is why we do not lock all
            # resolving objects for update, but instead only lock one object at a time. This
            # allows managers running in parallel to process different objects.
            data = Data.objects.select_for_update().get(pk=data.pk)
            if data.status != Data.STATUS_RESOLVING:
                # The object might have already been processed while waiting for the lock to be
                # obtained. In this case, skip the object.
                return

            dep_status = data.dependency_status()

            if dep_status == Data.STATUS_ERROR:
                data.status = Data.STATUS_ERROR
                data.process_error.append("One or more inputs have status ERROR")
                data.process_rc = 1
                data.save()
                if hasattr(data, "worker"):
                    data.worker.status = Worker.STATUS_ERROR_PREPARING
                    data.worker.save(update_fields=["status"])

                return

            elif dep_status != Data.STATUS_DONE:
                return

            run_in_executor = False
            if data.process.run:
                try:
                    # Check if execution engine is sound and evaluate workflow.
                    execution_engine_name = data.process.run.get("language", None)
                    execution_engine = self.get_execution_engine(execution_engine_name)
                    run_in_executor = execution_engine_name != "workflow"
                    if not run_in_executor:
                        execution_engine.evaluate(data)
                    else:
                        # Set allocated resources
                        resource_limits = data.get_resource_limits()
                        data.process_memory = resource_limits["memory"]
                        data.process_cores = resource_limits["cores"]

                except (ExecutionError, InvalidEngineError) as error:
                    data.status = Data.STATUS_ERROR
                    data.process_error.append(
                        "Error in process script: {}".format(error)
                    )
                    data.save()
                    if hasattr(data, "worker"):
                        data.worker.status = Worker.STATUS_ERROR_PREPARING
                        data.worker.save(update_fields=["status"])

                    return
            if data.status != Data.STATUS_DONE:
                # The data object may already be marked as done by the execution engine. In this
                # case we must not revert the status to STATUS_WAITING.
                data.status = Data.STATUS_WAITING
            data.save(render_name=True)

            # Actually run the object only if there was nothing with the
            # transaction and was not already evaluated.
            if run_in_executor:
                transaction.on_commit(
                    # Make sure the closure gets the right values here, since they're
                    # changed in the loop.
                    lambda d=data: self._data_execute(d)
                )

        logger.debug(
            __(
                "Manager processing data scan triggered by Data with id {}.",
                data_id,
            )
        )

        try:
            queryset = Data.objects.filter(status=Data.STATUS_RESOLVING)
            # Check if process is in the whitelist or blacklist. The blacklist has
            # priority.
            if self._processes_allow:
                queryset = queryset.filter(process__slug__in=self._processes_allow)
            if self._processes_ignore:
                queryset = queryset.exclude(process__slug__in=self._processes_ignore)

            if data_id is not None:
                # Scan only given data object and its children.
                queryset = queryset.filter(
                    Q(parents=data_id) | Q(id=data_id)
                ).distinct()

            for data in queryset:
                try:
                    with transaction.atomic():
                        process_data_object(data)

                        # All data objects created by the execution engine are commited after this
                        # point and may be processed by other managers running in parallel. At the
                        # same time, the lock for the current data object is released.
                except Exception as error:
                    logger.exception(
                        __(
                            "Unhandled exception in _data_scan while processing data object {}.",
                            data.pk,
                        )
                    )

                    # Unhandled error while processing a data object. We must set its
                    # status to STATUS_ERROR to prevent the object from being retried
                    # on next _data_scan run. We must perform this operation without
                    # using the Django ORM as using the ORM may be the reason the error
                    # occurred in the first place.
                    # Note that this has a side effect: since signals are not emitted,
                    # the data object is not processed and its children are not
                    # transitioned into the error state.
                    error_msg = "Internal error: {}".format(error)
                    process_error_field = Data._meta.get_field("process_error")
                    max_length = process_error_field.base_field.max_length
                    if len(error_msg) > max_length:
                        error_msg = error_msg[: max_length - 3] + "..."

                    try:
                        with connection.cursor() as cursor:
                            cursor.execute(
                                """
                                    UPDATE {table}
                                    SET
                                        status = %(status)s,
                                        process_error = process_error || (%(error)s)::varchar[]
                                    WHERE id = %(id)s
                                """.format(
                                    table=Data._meta.db_table
                                ),
                                {
                                    "status": Data.STATUS_ERROR,
                                    "error": [error_msg],
                                    "id": data.pk,
                                },
                            )
                        self.communicate(data_id=data.pk)
                    except Exception:
                        # If object's state cannot be changed due to some database-related
                        # issue, at least skip the object for this run.
                        logger.exception(
                            __(
                                "Unhandled exception in _data_scan while trying to emit error for {}.",
                                data.pk,
                            )
                        )

        except IntegrityError as exp:
            logger.error(__("IntegrityError in manager {}", exp))
            return

    def get_executor(self):
        """Return an executor instance."""
        return self.executor

    def get_expression_engine(self, name: str):
        """Return an expression engine instance."""
        try:
            return self.expression_engines[name]
        except KeyError:
            raise InvalidEngineError("Unsupported expression engine: {}".format(name))

    def get_execution_engine(self, name: str):
        """Return an execution engine instance."""
        try:
            return self.execution_engines[name]
        except KeyError:
            raise InvalidEngineError("Unsupported execution engine: {}".format(name))

    def load_executor(self, executor_name: str):
        """Load process executor."""
        executor_name = executor_name + ".prepare"
        module = import_module(executor_name)
        return module.FlowExecutorPreparer()  # type: ignore

    def load_expression_engines(self, engines: List[Union[dict, str]]):
        """Load expression engines."""
        return load_engines(self, "ExpressionEngine", "expression_engines", engines)

    def load_execution_engines(self, engines: List[Union[dict, str]]):
        """Load execution engines."""
        return load_engines(self, "ExecutionEngine", "execution_engines", engines)



""".. Ignore pydocstyle D400.

===============
Local Connector
===============

"""

import logging
import os
import subprocess

from resolwe.flow.managers.listener.listener import LISTENER_PUBLIC_KEY
from resolwe.flow.models import Data
from resolwe.storage import settings as storage_settings
from resolwe.utils import BraceMessage as __

from .base import BaseConnector

logger = logging.getLogger(__name__)


class Connector(BaseConnector):
    """Local connector for job execution."""

    def submit(self, data: Data, argv):
        """Run process locally.

        For details, see
        :meth:`~resolwe.flow.managers.workload_connectors.base.BaseConnector.submit`.
        """
        logger.debug(
            __(
                "Connector '{}' running for Data with id {} ({}).",
                self.__class__.__module__,
                data.id,
                repr(argv),
            )
        )
        process_environment = os.environ.copy()
        process_environment["LISTENER_PUBLIC_KEY"] = LISTENER_PUBLIC_KEY
        process_environment["CURVE_PRIVATE_KEY"] = data.worker.private_key
        process_environment["CURVE_PUBLIC_KEY"] = data.worker.public_key
        runtime_dir = storage_settings.FLOW_VOLUMES["runtime"]["config"]["path"]
        subprocess.Popen(
            argv, env=process_environment, cwd=runtime_dir, stdin=subprocess.DEVNULL
        ).wait()



"""Upload file process."""

from pathlib import Path

from resolwe.process import FileField, Process, SchedulingClass


class UploadFile(Process):
    """Upload file process."""

    slug = "upload-file"
    name = "File"
    process_type = "data:file"
    data_name = '{{ src.file|default("?") }}'
    version = "1.1.2"
    category = "Import"
    description = "Import any other file format such as a pdf."
    requirements = {
        "expression-engine": "jinja",
        "executor": {
            "docker": {
                "image": "public.ecr.aws/s4q6j6e8/resolwe/base:ubuntu-20.04-03022021",
            },
        },
        "resources": {
            "cores": 1,
            "memory": 1024,
            "network": True,
        },
    }
    scheduling_class = SchedulingClass.BATCH

    class Input:
        """Input parameters."""

        src = FileField(label="Input file")

    class Output:
        """Output parameters."""

        file = FileField(label="File")

    def run(self, inputs, outputs):
        """Upload file."""
        supported_suffixes = [".gz", ".bz2", ".zip", ".rar", ".7z"]
        input_file = Path(inputs.src.import_file())

        if Path(f"{input_file}.tar.gz").is_file():
            outputs.file = f"{input_file}.tar.gz"
        elif Path(inputs.src.path).suffix in supported_suffixes:
            outputs.file = f"{input_file}.gz"
        else:
            outputs.file = f"{input_file}"



""".. Ignore pydocstyle D400.

====================
Docker Flow Executor
====================

.. automodule:: resolwe.flow.executors.docker.run
.. automodule:: resolwe.flow.executors.docker.prepare

"""



"""Filters for Jinja expression engine."""

import copy
import json
import os

from django.conf import settings

from resolwe.flow.models import Data
from resolwe.flow.models.utils import hydrate_input_references, hydrate_input_uploads
from resolwe.flow.utils import dict_dot


def apply_filter_list(func, obj):
    """Apply `func` to list or tuple `obj` element-wise and directly otherwise."""
    if isinstance(obj, (list, tuple)):
        return [func(item) for item in obj]
    return func(obj)


def _get_data_attr(data, attr):
    """Get data object field."""
    if isinstance(data, dict):
        # `Data` object's id is hydrated as `__id` in expression engine
        data = data["__id"]

    data_obj = Data.objects.get(id=data)

    return getattr(data_obj, attr)


def name(data):
    """Return `name` of `Data`."""
    return apply_filter_list(lambda datum: _get_data_attr(datum, "name"), data)


def slug(data):
    """Return `slug` of `Data`."""
    return apply_filter_list(lambda datum: _get_data_attr(datum, "slug"), data)


def input_(data, field_path):
    """Return a hydrated value of the ``input`` field."""
    data_obj = Data.objects.get(id=data["__id"])

    inputs = copy.deepcopy(data_obj.input)
    # XXX: Optimize by hydrating only the required field (major refactoring).
    hydrate_input_references(inputs, data_obj.process.input_schema)
    hydrate_input_uploads(inputs, data_obj.process.input_schema)

    return dict_dot(inputs, field_path)


def id_(obj):
    """Return ``id`` key of dict."""
    return apply_filter_list(lambda item: item["__id"], obj)


def type_(obj):
    """Return ``type`` key of dict."""
    return apply_filter_list(lambda item: item["__type"], obj)


def basename(path):
    """Return the base name of pathname path."""
    return os.path.basename(path)


def dirname(path):
    """Return the base name of pathname path."""
    return os.path.dirname(path)


def subtype(basetype, supertype):
    """Check if ``basetype`` is a subtype of supertype."""
    return apply_filter_list(lambda item: item.startswith(supertype), basetype)


def yesno(value, true_value, false_value):
    """Return ``true_value`` if ``value`` evaluates to true and ``false_value`` otherwise."""
    return true_value if value else false_value


def data_by_slug(data_slug):
    """Return the primary key of a data object identified by the given slug."""
    return Data.objects.get(slug=data_slug).pk


def _get_hydrated_path(field):
    """Return HydratedPath object for file-type field."""
    # Get only file path if whole file object is given.
    if isinstance(field, str) and hasattr(field, "file_name"):
        # field is already actually a HydratedPath object
        return field

    if isinstance(field, dict) and "file" in field:
        hydrated_path = field["file"]

    if not hasattr(hydrated_path, "file_name"):
        raise TypeError("Filter argument must be a valid file-type field.")

    return hydrated_path


def get_url(field):
    """Return file's url based on base url set in settings."""
    hydrated_path = _get_hydrated_path(field)
    base_url = getattr(settings, "RESOLWE_HOST_URL", "localhost")
    return "{}/data/{}/{}".format(
        base_url, hydrated_path.data_id, hydrated_path.file_name
    )


def relative_path(field):
    """Return file's relative path."""
    hydrated_path = _get_hydrated_path(field)
    return hydrated_path.file_name


def descriptor(obj, path=""):
    """Return descriptor of given object.

    If ``path`` is specified, only the content on that path is
    returned.
    """
    if isinstance(obj, dict):
        # Current object is hydrated, so we need to get descriptor from
        # dict representation.
        desc = obj["__descriptor"]
    else:
        desc = obj.descriptor

    resp = dict_dot(desc, path)

    if isinstance(resp, list) or isinstance(resp, dict):
        return json.dumps(resp)

    return resp


def all_(obj):
    """Return True if all items in obj are true or if obj is empty."""
    return all(obj)


def any_(obj):
    """Return True if any item in obj is true. If obj is empty, return False."""
    return any(obj)


# A dictionary of filters that will be registered.
filters = {
    "name": name,
    "slug": slug,
    "input": input_,
    "id": id_,
    "type": type_,
    "basename": basename,
    "dirname": dirname,
    "subtype": subtype,
    "yesno": yesno,
    "data_by_slug": data_by_slug,
    "get_url": get_url,
    "relative_path": relative_path,
    "descriptor": descriptor,
    "all": all_,
    "any": any_,
}



# pylint: disable=missing-docstring
import asyncio
import logging
import os
import subprocess
import sys
import unittest
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path
from time import sleep, time
from unittest import mock
from unittest.mock import patch

import zmq
import zmq.asyncio
from django.conf import settings
from django.db import IntegrityError, connection
from django.test import override_settings

from resolwe.flow.executors.prepare import BaseFlowExecutorPreparer
from resolwe.flow.executors.socket_utils import Message, Response, ResponseStatus
from resolwe.flow.executors.startup_processing_container import (
    connect_to_communication_container,
    initialize_connections,
)
from resolwe.flow.executors.zeromq_utils import ZMQCommunicator
from resolwe.flow.managers.dispatcher import Manager
from resolwe.flow.managers.listener.listener import LISTENER_PUBLIC_KEY
from resolwe.flow.managers.listener.redis_cache import redis_cache
from resolwe.flow.models import Data, DataDependency, Process, Worker
from resolwe.flow.models.annotations import (
    AnnotationField,
    AnnotationGroup,
    AnnotationType,
)
from resolwe.flow.models.fields import ResolweSlugField
from resolwe.storage import settings as storage_settings
from resolwe.test import (
    ProcessTestCase,
    TestCase,
    tag_process,
    with_docker_executor,
    with_null_executor,
)

PROCESSES_DIR = os.path.join(os.path.dirname(__file__), "processes")
DESCRIPTORS_DIR = os.path.join(os.path.dirname(__file__), "descriptors")

logger = logging.getLogger(__name__)


class GetToolsTestCase(TestCase):
    @mock.patch("resolwe.flow.utils.apps")
    @mock.patch("resolwe.flow.utils.Path")
    @mock.patch("resolwe.flow.utils.settings")
    def test_get_tools_paths(self, settings_mock, path_mock, apps_mock):
        def path_mock_side_effect(*args):
            if args[0] == "/custom_tools":
                return mock.MagicMock(path=args[0])
            else:
                return app_config_path_mock

        path_mock.side_effect = path_mock_side_effect

        app_config1_mock = mock.MagicMock()
        app_config1_mock.name = "/resolwe/test_app1"
        app_config2_mock = mock.MagicMock()
        app_config2_mock.name = "/resolwe/test_app1"

        apps_mock.get_app_configs.return_value = [app_config1_mock, app_config2_mock]
        first_mock = mock.MagicMock(is_dir=lambda: False, path="/resolwe/test_app1")
        second_mock = mock.MagicMock(is_dir=lambda: True, path="/resolwe/test_app2")

        app_config_path_mock = mock.MagicMock()
        app_config_path_mock.__truediv__.side_effect = [first_mock, second_mock]

        settings_mock.RESOLWE_CUSTOM_TOOLS_PATHS = ["/custom_tools"]
        base_executor = BaseFlowExecutorPreparer()
        tools_list = [mock.path for mock in base_executor.get_tools_paths()]

        self.assertEqual(len(tools_list), 2)
        self.assertIn("/resolwe/test_app2", tools_list)
        self.assertIn("/custom_tools", tools_list)

    @mock.patch("resolwe.flow.utils.apps")
    @mock.patch("resolwe.flow.utils.settings")
    def test_not_list(self, settings_mock, apps_mock):
        apps_mock.get_app_configs.return_value = []
        settings_mock.RESOLWE_CUSTOM_TOOLS_PATHS = "/custom_tools"

        base_executor = BaseFlowExecutorPreparer()
        with self.assertRaisesRegex(KeyError, "setting must be a list"):
            base_executor.get_tools_paths()


class ManagerRunProcessTest(ProcessTestCase):
    def setUp(self):
        super().setUp()
        self._register_schemas(
            processes_paths=[PROCESSES_DIR], descriptors_paths=[DESCRIPTORS_DIR]
        )
        group = AnnotationGroup.objects.create(name="general", sort_order=1)
        self.species_field = AnnotationField.objects.create(
            name="species",
            group=group,
            sort_order=1,
            type=AnnotationType.STRING.value,
            vocabulary={"Valid": "Valid"},
        )
        AnnotationField.objects.create(
            name="description",
            group=group,
            sort_order=2,
            type=AnnotationType.STRING.value,
        )

    @tag_process("test-min")
    def test_minimal_process(self):
        self.run_process("test-min")

    @tag_process("test-min")
    @tag_process("test-workflow-simple")
    def test_restart(self):
        def restarted_object_sanity_check(
            original_values, restarted_values, additional_skip_keys=[]
        ):
            self.assertEqual(restarted_values["status"], Data.STATUS_DONE)
            self.assertEqual(original_values.keys(), restarted_values.keys())

            # These values should be both None or modified greater than original.
            for key in {"started", "finished", "modified", "scheduled"}:
                if restarted_values[key] is not None:
                    compare_value = original_values["finished"] or original_values[key]
                    self.assertGreater(restarted_values[key], compare_value)
                else:
                    self.assertEqual(restarted_values[key], original_values[key])

            skip_keys = {"started", "finished", "modified", "scheduled", "location_id"}
            skip_keys.update(additional_skip_keys)
            for key in original_values.keys():
                if key not in skip_keys:
                    self.assertEqual(
                        original_values[key],
                        restarted_values[key],
                        f"Key '{key}' does not match.",
                    )

        original = self.run_process("test-min", contributor=self.contributor)
        with self.assertRaises(
            RuntimeError, msg="Only data in status ERROR can be restarted, not OK."
        ):
            original.restart()

        original_values = Data.objects.filter(pk=original.pk).values()[0]

        # No need to actually save the status, it is checked on the instance.
        original.status = Data.STATUS_ERROR
        original.restart()

        restarted_values = Data.objects.filter(pk=original.pk).values()[0]
        restarted_object_sanity_check(original_values, restarted_values)

        # Now test the workflow duplication. The workflow creates two data objects
        # and the second one depends on the first one.
        workflow_data = self.run_process(
            "test-workflow-simple",
            {"param1": "world"},
            tags=["test-tag"],
            contributor=self.contributor,
        )
        workflow_data.refresh_from_db()
        step1_data = Data.objects.get(process__slug="test-example-1-simple")
        step2_data = Data.objects.get(process__slug="test-example-2-simple")

        workflow_dict = Data.objects.filter(pk=workflow_data.pk).values()[0]
        step1_dict = Data.objects.filter(pk=step1_data.pk).values()[0]
        step2_dict = Data.objects.filter(pk=step2_data.pk).values()[0]

        # First try to duplicate the entire workflow.
        workflow_data.status = Data.STATUS_ERROR
        workflow_data.restart()

        # Subprocess dependencies must be deleted.
        with self.assertRaises(Data.DoesNotExist):
            step1_data.refresh_from_db()
        with self.assertRaises(Data.DoesNotExist):
            step2_data.refresh_from_db()

        restarted_workflow_dict = Data.objects.filter(pk=workflow_data.pk).values()[0]
        restarted_step1_dict = Data.objects.filter(
            process__slug="test-example-1-simple"
        ).values()[0]
        restarted_step2_dict = Data.objects.filter(
            process__slug="test-example-2-simple"
        ).values()[0]
        # Output also differs (different step ids).
        restarted_object_sanity_check(
            workflow_dict, restarted_workflow_dict, ["output"]
        )
        # Ids will not match, since object is restarted.
        self.assertNotEqual(restarted_step1_dict["id"], step1_dict["id"])
        self.assertNotEqual(restarted_step2_dict["id"], step2_dict["id"])
        self.assertGreater(restarted_step1_dict["created"], step1_dict["created"])
        self.assertGreater(restarted_step2_dict["created"], step2_dict["created"])
        restarted_object_sanity_check(
            step1_dict, restarted_step1_dict, ["id", "permission_group_id", "created"]
        )
        restarted_object_sanity_check(
            step2_dict, restarted_step2_dict, ["id", "permission_group_id", "created"]
        )

        # Override resources on step1 and step2. Be careful: when restarting the entire
        # workflow the workflow will spawn new objects and its children will be restarted.
        Data.objects.all().delete()
        workflow_data = self.run_process(
            "test-workflow-simple",
            {"param1": "world"},
            tags=["test-tag"],
            contributor=self.contributor,
        )
        workflow_data.refresh_from_db()

        step1_data = Data.objects.get(process__slug="test-example-1-simple")
        step2_data = Data.objects.get(process__slug="test-example-2-simple")

        workflow_dict = Data.objects.filter(pk=workflow_data.pk).values()[0]
        step1_dict = Data.objects.filter(pk=step1_data.pk).values()[0]
        step2_dict = Data.objects.filter(pk=step2_data.pk).values()[0]

        # Duplicate the entire workflow, override resources.
        workflow_data.status = Data.STATUS_ERROR

        resource_overrides = {}
        resource_overrides[workflow_data.pk] = {"cores": 1}
        workflow_data.restart(resource_overrides=resource_overrides)

        # Subprocess dependencies must be deleted.
        with self.assertRaises(Data.DoesNotExist):
            step1_data.refresh_from_db()
        with self.assertRaises(Data.DoesNotExist):
            step2_data.refresh_from_db()

        restarted_workflow_dict = Data.objects.filter(pk=workflow_data.pk).values()[0]
        restarted_step1_dict = Data.objects.filter(
            process__slug="test-example-1-simple"
        ).values()[0]
        restarted_step2_dict = Data.objects.filter(
            process__slug="test-example-2-simple"
        ).values()[0]
        # Output also differs (different step ids).
        restarted_object_sanity_check(
            workflow_dict, restarted_workflow_dict, ["output", "process_resources"]
        )
        # Ids will not match, since object is restarted.
        self.assertNotEqual(restarted_step1_dict["id"], step1_dict["id"])
        self.assertNotEqual(restarted_step2_dict["id"], step2_dict["id"])
        self.assertGreater(restarted_step1_dict["created"], step1_dict["created"])
        self.assertGreater(restarted_step2_dict["created"], step2_dict["created"])
        restarted_object_sanity_check(
            step1_dict, restarted_step1_dict, ["id", "permission_group_id", "created"]
        )
        restarted_object_sanity_check(
            step2_dict, restarted_step2_dict, ["id", "permission_group_id", "created"]
        )
        self.assertEqual(restarted_workflow_dict["process_resources"], {"cores": 1})
        self.assertEqual(workflow_dict["process_resources"], {})

    @tag_process("test-annotate")
    def test_annotate(self):
        data = self.run_process("test-annotate")
        self.assertIsNotNone(data.entity)
        self.assertEqual(data.entity.annotations.count(), 1)
        self.assertAnnotation(data.entity, "general.species", "Valid")

    def test_router_handover(self):
        """Test router socket handover.

        When multiple clients with the same identity connect to the socket, the
        last one must replace the previous one.
        """
        listener_settings = getattr(settings, "FLOW_EXECUTOR", {}).get(
            "LISTENER_CONNECTION", {}
        )
        port = listener_settings.get("port", 53893)
        host = listener_settings.get("hosts", {}).get("local", "127.0.0.1")
        protocol = settings.FLOW_EXECUTOR.get("LISTENER_CONNECTION", {}).get(
            "protocol", "tcp"
        )
        logger = logging.getLogger(__name__)
        logger.handlers = []
        process = Process.objects.create(contributor=self.contributor)
        data = Data.objects.create(process=process, contributor=self.contributor)
        public_key, private_key = zmq.curve_keypair()
        Worker.objects.create(
            data=data,
            status=Worker.STATUS_PROCESSING,
            private_key=private_key,
            public_key=public_key,
        )

        async def send_single_message():
            """Open connection to listener and send single message."""
            connection_string = f"{protocol}://{host}:{port}"
            zmq_context = zmq.asyncio.Context.instance()
            zmq_socket = zmq_context.socket(zmq.DEALER)
            zmq_socket.curve_secretkey = private_key
            zmq_socket.curve_publickey = public_key
            zmq_socket.curve_serverkey = LISTENER_PUBLIC_KEY
            zmq_socket.setsockopt(zmq.IDENTITY, b"1")
            zmq_socket.connect(connection_string)
            communicator = ZMQCommunicator(
                zmq_socket, "init_container <-> listener", logger
            )
            async with communicator:
                await asyncio.ensure_future(
                    communicator.send_command(Message.command("update_status", "PP"))
                )

        async def send_two_messages():
            """Send two messages with the same identity.

            When handover is not working the second connection will be rejected
            and no answer received, triggering timeout and raising exception.
            """
            for _ in range(2):
                await send_single_message()

        return asyncio.new_event_loop().run_until_complete(send_two_messages())

    # @mock.patch("resolwe.flow.managers.workload_connectors.local.LISTENER_PUBLIC_KEY", b"0")
    def test_encryption(self):
        """Test encryption between the listener and the process.

        Verify that:
        - invalid listener key in the process results in the failure to communicate.
        - invalid process key results in the error in the listener.
        - the process with valid key can only process its data object.
        """
        listener_settings = getattr(settings, "FLOW_EXECUTOR", {}).get(
            "LISTENER_CONNECTION", {}
        )
        port = listener_settings.get("port", 53893)
        host = listener_settings.get("hosts", {}).get("local", "127.0.0.1")
        protocol = settings.FLOW_EXECUTOR.get("LISTENER_CONNECTION", {}).get(
            "protocol", "tcp"
        )
        process = Process.objects.create(contributor=self.contributor)
        data = Data.objects.create(process=process, contributor=self.contributor)
        another_data = Data.objects.create(
            process=process, contributor=self.contributor
        )
        public_key, private_key = zmq.curve_keypair()
        another_public_key, another_private_key = zmq.curve_keypair()
        fake_public_key, fake_private_key = zmq.curve_keypair()
        Worker.objects.create(
            data=data,
            status=Worker.STATUS_PROCESSING,
            private_key=private_key,
            public_key=public_key,
        )
        Worker.objects.create(
            data=another_data,
            status=Worker.STATUS_PROCESSING,
            private_key=another_private_key,
            public_key=another_public_key,
        )

        async def send_single_message(
            identity,
            public_key,
            private_key,
            listener_public_key,
            command=("update_status", "PP"),
        ) -> Response:
            """Open connection to listener and send single message."""
            connection_string = f"{protocol}://{host}:{port}"
            zmq_context = zmq.asyncio.Context.instance()
            zmq_socket = zmq_context.socket(zmq.DEALER)
            zmq_socket.curve_secretkey = private_key
            zmq_socket.curve_publickey = public_key
            zmq_socket.curve_serverkey = listener_public_key
            zmq_socket.setsockopt(zmq.IDENTITY, identity)
            zmq_socket.connect(connection_string)
            communicator = ZMQCommunicator(
                zmq_socket, "init_container <-> listener", logger
            )
            async with communicator:
                future = asyncio.ensure_future(
                    communicator.send_command(Message.command(*command))
                )
                return await asyncio.wait_for(future, 1)

        # Correct identity and keys.
        response = asyncio.new_event_loop().run_until_complete(
            send_single_message(
                f"{data.pk}".encode(), public_key, private_key, LISTENER_PUBLIC_KEY
            )
        )
        self.assertEqual(response.status, ResponseStatus.OK)

        # Fake listener identity.
        with self.assertRaises(asyncio.exceptions.TimeoutError):
            asyncio.new_event_loop().run_until_complete(
                send_single_message(
                    f"{data.pk}".encode(), public_key, private_key, fake_public_key
                )
            )

        # Non-matching key pair should not be able to establish a connection.
        with self.assertRaises(asyncio.exceptions.TimeoutError):
            asyncio.new_event_loop().run_until_complete(
                send_single_message(
                    f"{data.pk}".encode(),
                    fake_public_key,
                    private_key,
                    LISTENER_PUBLIC_KEY,
                )
            )
        with self.assertRaises(asyncio.exceptions.TimeoutError):
            asyncio.new_event_loop().run_until_complete(
                send_single_message(
                    f"{data.pk}".encode(),
                    public_key,
                    fake_private_key,
                    LISTENER_PUBLIC_KEY,
                )
            )

        # Fake client keys should be rejected.
        response = asyncio.new_event_loop().run_until_complete(
            send_single_message(
                f"{data.pk}".encode(),
                fake_public_key,
                fake_private_key,
                LISTENER_PUBLIC_KEY,
            )
        )
        not_authorized_error = (
            f"Client with key {fake_public_key!r} is not allowed to process the data "
            f"object with id {data.pk}."
        )
        self.assertEqual(response.status, ResponseStatus.ERROR)
        self.assertEqual(response.message_data, not_authorized_error)

        # The command 'liveness_probe' must be able to get a response even with a set
        # of keys with access to no data object.
        response = asyncio.new_event_loop().run_until_complete(
            send_single_message(
                b"liveness_probe",
                fake_public_key,
                fake_private_key,
                LISTENER_PUBLIC_KEY,
                command=("liveness_probe", ""),
            )
        )
        self.assertEqual(response.status, ResponseStatus.OK)

        # Non-matching keys and data id message must be rejected.
        response = asyncio.new_event_loop().run_until_complete(
            send_single_message(
                f"{another_data.pk}".encode(),
                public_key,
                private_key,
                LISTENER_PUBLIC_KEY,
            )
        )
        not_authorized_error = (
            f"Client with key {public_key!r} is not allowed to process the data "
            f"object with id {another_data.pk}."
        )
        self.assertEqual(response.status, ResponseStatus.ERROR)
        self.assertEqual(response.message_data, not_authorized_error)

        response = asyncio.new_event_loop().run_until_complete(
            send_single_message(
                f"{data.pk}".encode(),
                another_public_key,
                another_private_key,
                LISTENER_PUBLIC_KEY,
            )
        )
        not_authorized_error = (
            f"Client with key {another_public_key!r} is not allowed to process the "
            f"data object with id {data.pk}."
        )
        self.assertEqual(response.status, ResponseStatus.ERROR)
        self.assertEqual(response.message_data, not_authorized_error)

        # Correct identity and keys.
        response = asyncio.new_event_loop().run_until_complete(
            send_single_message(
                f"{data.pk}".encode(), public_key, private_key, LISTENER_PUBLIC_KEY
            )
        )
        self.assertEqual(response.status, ResponseStatus.OK)

        response = asyncio.new_event_loop().run_until_complete(
            send_single_message(
                f"{another_data.pk}".encode(),
                another_public_key,
                another_private_key,
                LISTENER_PUBLIC_KEY,
            )
        )
        self.assertEqual(response.status, ResponseStatus.OK)

    @tag_process("test-annotate-wrong-option")
    def test_annotate_wrong_option(self):
        data = self.run_process(
            "test-annotate-wrong-option", assert_status=Data.STATUS_ERROR
        )
        self.assertEqual(len(data.process_error), 1)
        self.assertIn(
            "The value 'Invalid' is not valid for the field general.species.",
            data.process_error[0],
        )

    @tag_process("test-annotate-wrong-type")
    def test_annotate_wrong_type(self):
        data = self.run_process(
            "test-annotate-wrong-type", assert_status=Data.STATUS_ERROR
        )
        self.assertEqual(len(data.process_error), 1)
        self.assertIn(
            "The value '15' is not of the expected type 'str'.", data.process_error[0]
        )

    @tag_process("test-annotate-missing-field")
    def test_annotate_missing_field(self):
        data = self.run_process(
            "test-annotate-missing-field", assert_status=Data.STATUS_ERROR
        )
        self.assertEqual(len(data.process_error), 1)
        self.assertIn("Field 'general.invalid' does not exist", data.process_error[0])

    @tag_process("test-annotate-no-entity")
    def test_annotate_no_entity(self):
        data = self.run_process(
            "test-annotate-no-entity", assert_status=Data.STATUS_ERROR
        )
        self.assertEqual(len(data.process_error), 1)
        self.assertIn("No entity to annotate", data.process_error[0])

    @tag_process("test-missing-file")
    def test_missing_file(self):
        self.run_process("test-missing-file", assert_status=Data.STATUS_ERROR)

        data = Data.objects.last()
        self.assertEqual(data.status, Data.STATUS_ERROR)

        self.assertEqual(len(data.process_error), 1)
        self.assertIn(
            "Output 'output' set to a missing file: 'i-dont-exist.zip'.",
            data.process_error[0],
        )

    @tag_process("test-spawn-new")
    def test_spawn(self):
        self.run_process("test-spawn-new", tags=["test-tag"])

        data = Data.objects.last()
        file_path = data.location.get_path(filename="foo.bar")
        self.assertEqual(data.output["saved_file"]["file"], "foo.bar")
        self.assertTrue(os.path.isfile(file_path))
        with open(file_path, "rt") as stream:
            self.assertEqual(stream.read(), "foo.bar\n")
        self.assertEqual(data.tags, ["test-tag"])
        parent_data = Data.objects.first()
        self.assertEqual(data.parents.count(), 1)
        self.assertEqual(data.parents.first(), parent_data)

        # Check correct dependency type is created.
        self.assertEqual(
            {d.kind for d in data.parents_dependency.all()},
            {DataDependency.KIND_SUBPROCESS},
        )

    @tag_process("test-spawn-missing-file")
    def test_spawn_missing_export(self):
        data = self.run_process(
            "test-spawn-missing-file", assert_status=Data.STATUS_ERROR
        )
        self.assertIn(
            "Error while preparing spawned Data objects", data.process_error[0]
        )

    @tag_process(
        "test-broken",
        "test-broken-invalid-execution-engine",
        "test-broken-invalid-expression-engine",
        "test-broken-data-name",
    )
    def test_broken(self):
        Process.objects.create(
            slug="test-broken-invalid-execution-engine",
            name="Test Process",
            contributor=self.contributor,
            type="data:test",
            version=1,
            run={
                "language": "invalid",
            },
        )

        self.run_process("test-broken", assert_status=Data.STATUS_ERROR)
        self.run_process(
            "test-broken-invalid-expression-engine", assert_status=Data.STATUS_ERROR
        )
        self.run_process(
            "test-broken-invalid-execution-engine", assert_status=Data.STATUS_ERROR
        )

        # If evaluation of data_name template fails, the process should not abort as the
        # template may be evaluatable later when the process completes.
        self.run_process("test-broken-data-name")

    @tag_process("test-broken-invalide-storage")
    def test_invalid_storage_file(self):
        data = self.run_process(
            "test-broken-invalide-storage", assert_status=Data.STATUS_ERROR
        )

        self.assertEqual(data.status, Data.STATUS_ERROR)
        self.assertIn("must be a valid JSON, current: 1a", data.process_error[0])

    @tag_process("test-workflow-1")
    def test_workflow(self):
        workflow_data = self.run_process(
            "test-workflow-1", {"param1": "world"}, tags=["test-tag"]
        )
        workflow_data.refresh_from_db()
        step1_data = Data.objects.get(process__slug="test-example-1")
        step2_data = Data.objects.get(process__slug="test-example-2")

        # Workflow should output indices of all data objects, in order.
        self.assertEqual(workflow_data.output["steps"], [step1_data.pk, step2_data.pk])

        # Steps should execute with the correct variables.

        self.assertEqual(step1_data.input["param1"], "world")
        self.assertEqual(step1_data.input["param2"], True)
        self.assertEqual(step1_data.output["out1"], "hello world")
        self.assertEqual(step1_data.tags, ["test-tag"])
        self.assertEqual(step2_data.input["param1"], step1_data.pk)
        self.assertEqual(step2_data.input["param2"]["a"], step1_data.pk)
        self.assertEqual(step2_data.input["param2"]["b"], "hello")
        self.assertEqual(step2_data.output["out1"], "simon says: hello world")
        self.assertEqual(step2_data.tags, ["test-tag"])

        self.assertEqual(step1_data.parents.count(), 1)
        self.assertEqual(step1_data.parents.first(), workflow_data)
        self.assertEqual(step2_data.parents.count(), 2)
        self.assertCountEqual(step2_data.parents.all(), [workflow_data, step1_data])

        # Check correct dependency type is created.
        self.assertEqual(
            {d.kind for d in step1_data.parents_dependency.all()},
            {DataDependency.KIND_SUBPROCESS},
        )
        self.assertEqual(
            {d.kind for d in step2_data.parents_dependency.all()},
            {DataDependency.KIND_SUBPROCESS, DataDependency.KIND_IO},
        )

    @tag_process("test-workflow-2")
    def test_workflow_entity(self):
        with self.preparation_stage():
            # `self.run_process` adds input data in self.collection
            input_data = self.run_process("test-example-3")

        workflow = self.run_process("test-workflow-2", {"data1": input_data.pk})

        # Check that workflow results are created.
        self.assertEqual(
            Data.objects.filter(pk__in=workflow.output["steps"]).count(), 1
        )
        # There is a workflow and two "normal" data objects
        self.assertEqual(Data.objects.all().count(), 3)

    @with_docker_executor
    @tag_process("test-docker")
    def test_run_in_docker(self):
        data = self.run_process("test-docker")
        self.assertEqual(data.output["result"], "OK")

    @patch.object(Manager, "communicate")
    def test_slug_colision(self, communicate_mock):
        """Test slug colision resolution.

        Use mocking to force slug colision.

        This checks for a bug in permission architecture that caused exception
        due to missing permission_group when slug colision occured.
        """

        async def dummy_coroutine(*args, **kwargs):
            """Dummy coroutine."""
            pass

        def pre_save(instance, add):
            """Set the current slug or raise exception."""
            current_slug = slugs.pop(0)

            if isinstance(current_slug, Exception):
                raise current_slug

            instance.slug = current_slug
            return current_slug

        expected_slug = "my_data_slug"
        slugs = [IntegrityError("flow_data_slug"), expected_slug]
        process = Process.objects.create(contributor=self.contributor)

        # INFO: Remove this line and the coroutine above after we stop
        # supporting Python <= 3.8. Starting with Python 3.8 Mock objects can
        # be awaited.
        communicate_mock.side_effect = dummy_coroutine

        with patch.object(ResolweSlugField, "pre_save", side_effect=pre_save):
            data = Data.objects.create(process=process, contributor=self.contributor)
        data.refresh_from_db()
        self.assertEqual(data.slug, expected_slug)
        self.assertTrue(data.pk is not None)
        self.assertTrue(data.permission_group is not None)

    @with_docker_executor
    def test_terminate_worker(self):
        process = Process.objects.get(slug="test-terminate")

        def get_status():
            return (
                Data.objects.filter(name="Test terminate")
                .values_list("status", flat=True)
                .first()
            )

        def wait_status_change(expected_statuses, timeout=20):
            """Wait for the data object to change status.

            :attr timeout: how many seconds to wait for the status change.

            :attr expected_statuses: the set of expected statuses to wait for.

            :returns: the last known status.
            """
            start_time = time()
            status = None
            while time() - start_time < timeout and status not in expected_statuses:
                sleep(0.1)
                status = get_status()
            return status

        def check_processing():
            """Start processing data object by saving it."""
            # Wait up to 20s for process to start.
            wait_status_change([Data.STATUS_PROCESSING])
            data = Data.objects.get(name="Test terminate")
            # Terminate the worker and wait for the process to complete. In the
            # worst case it will take 60 seconds for the process to complete.
            logger.debug("test_terminate_worker terminating data '%s'.", data.pk)
            data.worker.terminate()
            wait_status_change([Data.STATUS_ERROR, Data.STATUS_DONE])
            connection.close()

        start_time = time()
        with ThreadPoolExecutor(max_workers=2) as executor:
            # Start the check_processing thread.
            executor.submit(check_processing)
            # Start processing data object.
            data = Data.objects.create(
                name="Test terminate", contributor=self.contributor, process=process
            )

        # To mark termination as a success it should take less than 50 seconds.
        # The uninterrupted process would run for 60 seconds.
        data.refresh_from_db()
        self.assertEqual(data.worker.status, Worker.STATUS_COMPLETED)
        self.assertEqual(data.status, Data.STATUS_ERROR)
        self.assertEqual(data.process_error[0], "Processing was cancelled.")
        self.assertLess(time() - start_time, 50)

    @with_docker_executor
    @tag_process("test-requirements-docker")
    def test_executor_requirements(self):
        data = self.run_process("test-requirements-docker")
        self.assertEqual(data.output["result"], "OK")

    @with_docker_executor
    @tag_process("test-docker-uid-gid")
    def test_docker_uid_gid(self):
        data = self.run_process("test-docker-uid-gid")
        self.assertEqual(data.output["result"], "OK")

    @unittest.skip("Null executor test currently not working.")
    @with_null_executor
    @tag_process("test-save-number")
    def test_null_executor(self):
        data = self.run_process(
            "test-save-number", {"number": 19}, assert_status=Data.STATUS_WAITING
        )
        self.assertEqual(data.input["number"], 19)
        self.assertEqual(data.output, {})

    @unittest.skipIf(
        os.environ.get("GITHUB_ACTIONS", "") == "true", "Fails on Github Actions"
    )
    @with_docker_executor
    @tag_process("test-memory-resource-alloc", "test-memory-resource-noalloc")
    def test_memory_resource(self):
        # This process should be terminated due to too much memory usage.
        self.run_process("test-memory-resource-alloc", assert_status=Data.STATUS_ERROR)
        # This process should run normally (honors the limits).
        self.run_process("test-memory-resource-noalloc")

    @with_docker_executor
    @tag_process("test-cpu-resource-1core", "test-cpu-resource-2core")
    def test_cpu_resource(self):
        # Currently there is no good way to test this reliably, so we just check if the
        # resource limit specification still makes the process run.
        self.run_process("test-cpu-resource-1core")
        self.run_process("test-cpu-resource-2core")

    @with_docker_executor
    @tag_process(
        "test-network-resource-enabled",
        "test-network-resource-disabled",
        "test-network-resource-policy",
    )
    def test_network_resource(self):
        self.run_process("test-network-resource-enabled")
        self.run_process(
            "test-network-resource-disabled", assert_status=Data.STATUS_ERROR
        )
        self.run_process(
            "test-network-resource-policy", assert_status=Data.STATUS_ERROR
        )

    @with_docker_executor
    @override_settings(FLOW_PROCESS_RESOURCE_DEFAULTS={"cpu_time_interactive": 1})
    @tag_process(
        "test-scheduling-class-interactive-ok",
        "test-scheduling-class-interactive-fail",
        "test-scheduling-class-batch",
    )
    def test_scheduling_class(self):
        self.run_process("test-scheduling-class-interactive-ok")
        self.run_process(
            "test-scheduling-class-interactive-fail", assert_status=Data.STATUS_ERROR
        )
        self.run_process("test-scheduling-class-batch")

    @with_docker_executor
    @tag_process("test-save-number")
    def test_executor_fs_lock(self):
        # First, run the process normaly. Do not autoremove container to check
        # if name clash with init contaner is avoided.
        with self.settings(FLOW_DOCKER_AUTOREMOVE=False):
            data = self.run_process("test-save-number", {"number": 42})
        # Make sure that process was successfully ran first time.
        self.assertEqual(data.output["number"], 42)
        data.output = {}
        data.save()

        file = Path(data.location.get_path(filename="temporary_file_do_not_purge.txt"))
        self.assertFalse(file.exists())
        file.touch()

        listener_settings = getattr(settings, "FLOW_EXECUTOR", {}).get(
            "LISTENER_CONNECTION", {}
        )
        port = listener_settings.get("port", 53893)
        hosts = settings.COMMUNICATION_CONTAINER_LISTENER_CONNECTION
        host = hosts.get("local", next(iter(hosts.values())))
        protocol = settings.FLOW_EXECUTOR.get("LISTENER_CONNECTION", {}).get(
            "protocol", "tcp"
        )

        # Set the status or listener will imediatelly respond with error
        # status. Make sure to also clear Redis cache for the data object.
        data.worker.status = Worker.STATUS_PROCESSING
        data.worker.save()
        data.status = Data.STATUS_PROCESSING
        data.save()
        redis_cache.clear(Data, (data.pk,))

        process_environment = os.environ.copy()
        process_environment["LISTENER_PUBLIC_KEY"] = LISTENER_PUBLIC_KEY
        process_environment["CURVE_PRIVATE_KEY"] = data.worker.private_key
        process_environment["CURVE_PUBLIC_KEY"] = data.worker.public_key

        process = subprocess.run(
            [
                "python",
                "-m",
                "executors",
                ".docker",
                str(data.id),
                host,
                str(port),
                protocol,
            ],
            cwd=storage_settings.FLOW_VOLUMES["runtime"]["config"]["path"],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            timeout=30,
            env=process_environment,
        )
        self.assertEqual(process.returncode, 0, f"The stderr was: '{process.stderr}'.")
        data.refresh_from_db()
        self.assertEqual(data.output, {})
        self.assertEqual(data.status, Data.STATUS_PROCESSING)
        self.assertEqual(data.process_error, [])
        self.assertEqual(data.process_info, [])
        # Check that temporary file was not deleted.
        self.assertTrue(file.exists())

    @mock.patch(
        "resolwe.flow.executors.startup_processing_container.asyncio.open_unix_connection"
    )
    def test_processing_communication_connection_ok(self, open_mock):
        """Test connection under normal conditions."""

        async def main_test():
            """Main test coroutine."""
            await asyncio.wait_for(
                connect_to_communication_container("anypath"), timeout=1
            )

        async def readline():
            """Return PING."""
            return b"PING"

        async def reader_writer():
            """Return reader-writer tuple mock."""
            return (mock.Mock(readline=readline), None)

        reader_mock = mock.MagicMock()
        reader_mock.readline = readline
        if sys.version_info >= (3, 8):
            open_mock.return_value = (mock.Mock(readline=readline), None)
        else:
            open_mock.return_value = reader_writer()
        loop = asyncio.new_event_loop()
        loop.run_until_complete(main_test())

    @mock.patch(
        "resolwe.flow.executors.startup_processing_container.asyncio.open_unix_connection"
    )
    def test_processing_communication_connection_ok_retry(self, open_mock):
        """Test connection if first attempt fails."""

        async def main_test():
            """Main test coroutine."""
            await asyncio.wait_for(
                connect_to_communication_container("anypath"), timeout=3
            )

        async def readline_ok():
            """Return PING."""
            return b"PING"

        async def readline_error():
            """Return ERROR."""
            return b"ERROR"

        async def reader_writer(coroutine):
            """Return reader-writer tuple mock."""
            return (mock.Mock(readline=coroutine), None)

        reader_mock = mock.MagicMock()
        reader_mock.readline.side_effect = [readline_error(), readline_ok()]
        if sys.version_info >= (3, 8):
            open_mock.return_value = (reader_mock, None)
        else:
            open_mock.side_effect = [
                reader_writer(readline_error),
                reader_writer(readline_ok),
            ]
        loop = asyncio.new_event_loop()
        loop.run_until_complete(main_test())

    @mock.patch(
        "resolwe.flow.executors.startup_processing_container.asyncio.open_unix_connection"
    )
    @mock.patch(
        "resolwe.flow.executors.startup_processing_container.constants.CONTAINER_TIMEOUT",
        2,
    )
    def test_processing_communication_connection_fail(self, open_mock):
        """Test function behaviour when connection fails."""

        async def main_test():
            """Main test coroutine."""
            current_loop = asyncio.get_event_loop()
            await initialize_connections(current_loop)

        async def readline_error():
            """Return ERROR."""
            return b"ERROR"

        async def reader_writer():
            """Return reader-writer tuple mock."""
            return (mock.Mock(readline=readline_error), None)

        reader_mock = mock.MagicMock()
        reader_mock.readline.side_effect = [readline_error(), readline_error()]
        if sys.version_info >= (3, 8):
            open_mock.return_value = (reader_mock, None)
        else:
            open_mock.side_effect = [reader_writer(), reader_writer()]
        loop = asyncio.new_event_loop()
        with self.assertRaisesRegex(
            ConnectionError, "Timeout connecting to communication container."
        ):
            loop.run_until_complete(main_test())

    @mock.patch(
        "resolwe.flow.executors.startup_processing_container.asyncio.open_unix_connection"
    )
    @mock.patch("resolwe.flow.executors.startup_processing_container.socket")
    @mock.patch(
        "resolwe.flow.executors.startup_processing_container.asyncio.start_unix_server"
    )
    def test_processing_communicate(self, server_mock, socket_mock, open_mock):
        """Test that communicate is called."""

        async def main_test():
            """Main test coroutine."""
            await asyncio.wait_for(
                connect_to_communication_container("anypath"), timeout=1
            )

        async def readline():
            """Return PING."""
            return b"PING"

        async def reader_writer():
            """Return reader-writer tuple mock."""
            return (mock.Mock(readline=readline), None)

        socket_mock.socket.return_value.connect.side_effect = None
        if sys.version_info >= (3, 8):
            open_mock.return_value = (mock.Mock(readline=readline), None)
        else:
            open_mock.return_value = reader_writer()
        # protocol_mock.return_value.communicate.side_effect = ValueError("test")
        loop = asyncio.new_event_loop()
        loop.run_until_complete(main_test())



# pylint: disable=missing-docstring
import os

from django.test import override_settings
from django.utils.crypto import get_random_string
from django.utils.text import slugify

from resolwe.flow.models import Process
from resolwe.test import ProcessTestCase


class PurgeTestFieldsMixin:
    """
    This class contains tests, which validate each field individually. It is used
    to perform basic unit tests (on simulated data, without running any processors)
    as well as more complex end-to-end tests (by running processors). Having them
    in this mixin makes sure that both, unit and e2e tests, are run for each field.
    """

    def test_basic_file(self):
        self.assertFieldWorks(
            "basic:file:",
            field_value={"file": "but_not_this"},
            script_setup="touch remove_this but_not_this",
            script_save="re-save-file sample but_not_this",
            removed=["remove_this"],
            not_removed=["but_not_this"],
        )

    def test_basic_file_specialization(self):
        self.assertFieldWorks(
            "basic:file:html:",
            field_value={"file": "but_not_this"},
            script_setup="touch remove_this but_not_this",
            script_save="re-save-file sample but_not_this",
            removed=["remove_this"],
            not_removed=["but_not_this"],
        )

    def test_basic_file_list(self):
        self.assertFieldWorks(
            "list:basic:file:",
            field_value=[{"file": "but_not_this"}, {"file": "and_not_this"}],
            script_setup="touch remove_this but_not_this and_not_this",
            script_save="re-save-file-list sample but_not_this and_not_this",
            removed=["remove_this"],
            not_removed=["but_not_this", "and_not_this"],
        )

    def test_basic_dir(self):
        self.assertFieldWorks(
            "basic:dir:",
            field_value={"dir": "but_not_this"},
            script_setup="""
mkdir remove_this but_not_this
touch remove_this/a remove_this/b remove_this/c
touch but_not_this/a but_not_this/b but_not_this/c
""",
            script_save="re-save-dir sample but_not_this",
            removed=["remove_this/", "remove_this/a", "remove_this/b", "remove_this/c"],
            not_removed=["but_not_this/a", "but_not_this/b", "but_not_this/c"],
        )

    def test_basic_dir_list(self):
        self.assertFieldWorks(
            "list:basic:dir:",
            field_value=[{"dir": "but_not_this"}, {"dir": "and_not_this"}],
            script_setup="""
mkdir remove_this but_not_this and_not_this
touch remove_this/a remove_this/b remove_this/c
touch but_not_this/a but_not_this/b but_not_this/c
touch and_not_this/a and_not_this/b and_not_this/c
""",
            script_save="re-save-dir-list sample but_not_this and_not_this",
            removed=["remove_this/", "remove_this/a", "remove_this/b", "remove_this/c"],
            not_removed=[
                "but_not_this/a",
                "but_not_this/b",
                "but_not_this/c",
                "and_not_this/a",
                "and_not_this/b",
                "and_not_this/c",
            ],
        )


@override_settings(TEST_PROCESS_REQUIRE_TAGS=False)  # Test uses dynamic processes.
class PurgeE2ETest(PurgeTestFieldsMixin, ProcessTestCase):
    def create_and_run_processor(self, processor, **kwargs):
        processor_slug = slugify(get_random_string(6))
        Process.objects.create(
            slug=processor_slug,
            name="Test Purge Process",
            contributor=self.admin,
            type="data:test",
            version=1,
            **processor,
        )
        data = self.run_process(processor_slug, **kwargs)
        return data

    def assertFilesRemoved(self, data, *files):
        for name in files:
            path = data.location.get_path(filename=name)
            self.assertFalse(os.path.isfile(path), msg=path)

    def assertFilesNotRemoved(self, data, *files):
        for name in files:
            path = data.location.get_path(filename=name)
            self.assertTrue(os.path.isfile(path), msg=path)

    def test_complex_processor(self):
        data = self.create_and_run_processor(
            processor=dict(
                input_schema=[],
                output_schema=[
                    {
                        "name": "sample_file",
                        "label": "Sample output file",
                        "type": "basic:file:",
                    },
                    {
                        "name": "another_sample_file",
                        "label": "Another sample output file",
                        "type": "basic:file:",
                    },
                    {
                        "name": "sample_dir",
                        "label": "Sample output directory",
                        "type": "basic:dir:",
                    },
                    {
                        "name": "sample_file_list",
                        "label": "Sample list of output files",
                        "type": "list:basic:file:",
                    },
                    {
                        "name": "sample_dir_list",
                        "label": "Sample list of output directories",
                        "type": "list:basic:dir:",
                    },
                ],
                run={
                    "language": "bash",
                    "program": """
touch these files should be removed
touch this_file_should_stay
mkdir -p should_stay/
touch should_stay/file
mkdir -p directory/should/be
touch directory/should/be/removed
touch directory/should/be/removed2
mkdir -p stay/directory
touch stay/directory/file1
touch stay/directory/file2
touch entry1 entry2 entry3
mkdir dir1 dir2 dir3
touch dir1/a dir2/b dir3/c dir3/d
touch ref1 ref2 ref3 ref4
mkdir refs
touch refs/a refs/b

re-save-file another_sample_file should_stay/file
re-save-file sample_file this_file_should_stay ref3
re-save-dir sample_dir stay ref4
re-save-file-list sample_file_list entry1 entry2:ref1 entry3:refs
re-save-dir-list sample_dir_list dir1:ref2 dir2 dir3
""",
                },
            )
        )

        self.assertFilesRemoved(
            data, "these", "files", "should", "be", "removed", "directory"
        )
        self.assertFilesNotRemoved(
            data,
            "this_file_should_stay",
            "stay/directory/file1",
            "stay/directory/file2",
            "entry1",
            "entry2",
            "entry3",
            "dir1/a",
            "dir2/b",
            "dir3/c",
            "dir3/d",
            "ref1",
            "ref2",
            "ref3",
            "ref4",
            "refs/a",
            "refs/b",
            "should_stay/file",
        )

    def assertFieldWorks(
        self, field_type, field_value, script_setup, script_save, removed, not_removed
    ):
        """
        Checks that a field is handled correctly when running a processor, which
        uses the field.
        """

        field_schema = {"name": "sample", "label": "Sample output", "type": field_type}

        # Test output.
        data = self.create_and_run_processor(
            processor=dict(
                input_schema=[],
                output_schema=[field_schema],
                run={"language": "bash", "program": script_setup + "\n" + script_save},
            )
        )
        self.assertFilesRemoved(data, *removed)
        self.assertFilesNotRemoved(data, *not_removed)

        # Test descriptor.
        # descriptor_schema = DescriptorSchema.objects.create(
        #     slug=get_random_string(6), contributor=self.admin, schema=[field_schema]
        # )
        # data = self.create_and_run_processor(
        #     processor=dict(
        #         input_schema=[],
        #         output_schema=[],
        #         run={"language": "bash", "program": script_setup},
        #     ),
        #     descriptor_schema=descriptor_schema,
        #     descriptor={"sample": field_value},
        # )

        # self.assertFilesRemoved(data, *removed)
        # self.assertFilesNotRemoved(data, *not_removed)



"""Storage manager."""

import logging
from typing import Optional

from django.db import transaction
from django.utils.timezone import now

from resolwe.storage.connectors import connectors
from resolwe.storage.models import AccessLog, FileStorage, StorageLocation
from resolwe.utils import BraceMessage as __

logger = logging.getLogger(__name__)


class Manager:
    """Storage manager."""

    def _lock_file_storage(self, file_storage_id: int) -> Optional[FileStorage]:
        """Lock file storage for processing and return it.

        :returns: the file storage object or None if locking fails.
        """
        return (
            FileStorage.objects.filter(id=file_storage_id)
            .select_for_update(skip_locked=True)
            .first()
        )

    def delete_single_location(self, file_storage: FileStorage, connector_name: str):
        """Delete given storage location."""
        delete_location = file_storage.storage_locations.filter(
            connector_name=connector_name
        ).first()
        # Some other storage manager might have already processed this storage
        # location before we did.
        if not delete_location:
            return
        logger.info(__("Deleting {} ({}).", file_storage, connector_name))
        delete_location.delete()

    def process_delete(self):
        """Delete storage locations."""
        for connector_name in connectors:
            logger.info(__("Deleting locations from {}.", connector_name))
            for file_storage in StorageLocation.objects.to_delete(
                connector_name
            ).iterator():
                with transaction.atomic():
                    file_storage = self._lock_file_storage(file_storage.id)
                    if file_storage is None:
                        continue
                    try:
                        self.delete_single_location(file_storage, connector_name)
                    except Exception:
                        logger.exception(
                            "Error deleting data from StorageLocation instance",
                        )

    def copy_single_location(self, file_storage: FileStorage, connector_name: str):
        """Copy given location to a given connector."""

        logger.info(__("Copying {} to {}.", file_storage, connector_name))
        storage_location = file_storage.default_storage_location
        assert storage_location is not None
        from_connector = connectors[storage_location.connector_name]

        access_log = None
        try:
            new_storage_location = StorageLocation.all_objects.get_or_create(
                file_storage=file_storage,
                url=storage_location.url,
                connector_name=connector_name,
                defaults={"status": StorageLocation.STATUS_UPLOADING},
            )[0]
            # Some other storage manager might have already processed this storage
            # location before we did.
            if new_storage_location.status == StorageLocation.STATUS_DONE:
                return

            access_log = AccessLog.objects.create(
                storage_location=storage_location,
                reason="Manager data transfer",
            )
            storage_location.transfer_data(new_storage_location)
            new_storage_location.status = StorageLocation.STATUS_DONE
            new_storage_location.save()
        except Exception:
            logger.exception(
                "Error transfering data",
                extra={
                    "file_storage_id": file_storage.pk,
                    "from_connector": from_connector.name,
                    "to_connector": connector_name,
                },
            )
        finally:
            if access_log is not None:
                access_log.finished = now()
                access_log.save()

    def process_copy(self):
        """Copy location to all applicable connectors."""
        for connector_name in connectors:
            logger.info(__("Copying locations to {}", connector_name))
            for file_storage in StorageLocation.objects.to_copy(
                connector_name
            ).iterator():
                with transaction.atomic():
                    file_storage = self._lock_file_storage(file_storage.id)
                    if file_storage is None:
                        continue
                    self.copy_single_location(file_storage, connector_name)

    def process(self):
        """Process all FileStorage objects."""
        logger.info("Starting storage manager copy run.")
        self.process_copy()
        logger.info("Starting storage manager delete run.")
        self.process_delete()
        logger.info("Storage manager run completed.")



""".. Ignore pydocstyle D400.

===========
Flow Models
===========

Base Model
==========

Base model for all other models.

.. autoclass:: resolwe.flow.models.base.BaseModel
    :members:

Collection Model
================

Postgres ORM model for the organization of collections.

.. autoclass:: resolwe.flow.models.collection.BaseCollection
    :members:

.. autoclass:: resolwe.flow.models.Collection
    :members:

Data model
==========

Postgres ORM model for keeping the data structured.

.. autoclass:: resolwe.flow.models.Data
    :members:

.. autoclass:: resolwe.flow.models.DataDependency
    :members:

Entity–relationship model
=========================

Postgres ORM to define the entity–relationship model that describes how
data objects are related in a specific domain.

.. autoclass:: resolwe.flow.models.Entity
    :members:

.. autoclass:: resolwe.flow.models.Relation
    :members:

.. autoclass:: resolwe.flow.models.RelationType
    :members:

DescriptorSchema model
======================

Postgres ORM model for storing descriptors.

.. autoclass:: resolwe.flow.models.DescriptorSchema
    :members:

Process model
=============

Postgres ORM model for storing processes.

.. autoclass:: resolwe.flow.models.Process
    :members:

Storage model
=============

Postgres ORM model for storing JSON.

.. autoclass:: resolwe.flow.models.Storage
    :members:

Secret model
============

Postgres ORM model for storing secrets.

.. autoclass:: resolwe.flow.models.Secret

ProcessMigrationHistory model
=============================

Postgres ORM model for storing proces migration history.

.. autoclass:: resolwe.flow.models.ProcessMigrationHistory
    :members:

DataMigrationHistory model
==========================

Postgres ORM model for storing data migration history.

.. autoclass:: resolwe.flow.models.DataMigrationHistory
    :members:

AnnotationGroup model
==========================

Postgres ORM model for storing annotation group data.

.. autoclass:: resolwe.flow.models.annotations.AnnotationGroup
    :members:

AnnotationField model
==========================

Postgres ORM model for storing annotation field data.

.. autoclass:: resolwe.flow.models.annotations.AnnotationField
    :members:

AnnotationPreset model
==========================

Postgres ORM model for storing annotation presets data.

.. autoclass:: resolwe.flow.models.annotations.AnnotationPreset
    :members:

AnnotationValue model
==========================

Postgres ORM model for storing annotation values.

.. autoclass:: resolwe.flow.models.annotations.AnnotationValue
    :members:

History model
=============

Basic model for history tracking.

.. autoclass:: resolwe.flow.models.history.History
    :members:

DataHistory model
=================

Model for tracking changes in Data objects.

.. autoclass:: resolwe.flow.models.history.DataHistory
    :members:

    
CollectionHistory model
=======================

Model for tracking changes in Collection objects.

.. autoclass:: resolwe.flow.models.history.CollectionHistory
    :members:

TrackChange model
=================

Basic model to track changes to particular model fields.

.. autoclass:: resolwe.flow.models.history.TrackChange
    :members:

DataSlugChange model
====================

Track data slug changes.

.. autoclass:: resolwe.flow.models.history.DataSlugChange
    :members:

CollectionSlugChange model
==========================

Track collection slug changes.

.. autoclass:: resolwe.flow.models.history.CollectionSlugChange
    :members:

DataNameChange model
====================

Track data name changes.

.. autoclass:: resolwe.flow.models.history.DataNameChange
    :members:

CollectionNameChange model
==========================

Track collection name changes.

.. autoclass:: resolwe.flow.models.history.CollectionNameChange
    :members:

    
CollectionChange model
======================

Track collection changes.

.. autoclass:: resolwe.flow.models.history.CollectionChange
    :members:

SizeChange model
================

Track size changes.

.. autoclass:: resolwe.flow.models.history.SizeChange
    :members:

ProcessingHistory model
=======================

Track processing data objects.

.. autoclass:: resolwe.flow.models.history.ProcessingHistory
    :members:


Utility functions
=================

.. autofunction:: resolwe.flow.models.utils.duplicate.bulk_duplicate



"""

from .annotations import (
    AnnotationField,
    AnnotationGroup,
    AnnotationPreset,
    AnnotationValue,
)
from .collection import Collection
from .data import Data, DataDependency
from .descriptor import DescriptorSchema
from .entity import Entity, Relation, RelationPartition, RelationType
from .history import (
    CollectionHistory,
    CollectionNameChange,
    CollectionSlugChange,
    DataHistory,
    DataNameChange,
    DataSlugChange,
)
from .migrations import DataMigrationHistory, ProcessMigrationHistory
from .process import Process
from .secret import Secret
from .storage import Storage
from .worker import Worker

__all__ = (
    "AnnotationGroup",
    "AnnotationPreset",
    "AnnotationField",
    "AnnotationValue",
    "Collection",
    "CollectionHistory",
    "CollectionNameChange",
    "CollectionSlugChange",
    "Data",
    "DataHistory",
    "DataDependency",
    "DataMigrationHistory",
    "DataNameChange",
    "DataSlugChange",
    "DescriptorSchema",
    "Entity",
    "Process",
    "ProcessMigrationHistory",
    "Relation",
    "RelationPartition",
    "RelationType",
    "Secret",
    "Storage",
    "Worker",
)



""".. Ignore pydocstyle D400.

===============
Slurm Connector
===============

"""

import logging
import os
import shlex
import subprocess

from django.conf import settings

from resolwe.flow.models import Data
from resolwe.storage import settings as storage_settings
from resolwe.utils import BraceMessage as __

from .base import BaseConnector

logger = logging.getLogger(__name__)

# We add this much to the memory limit to account for executor overhead,
# since the executor is running in the same environment as the process.
EXECUTOR_MEMORY_OVERHEAD = 200


class Connector(BaseConnector):
    """Slurm-based connector for job execution."""

    def submit(self, data: Data, argv):
        """Run process with SLURM.

        For details, see
        :meth:`~resolwe.flow.managers.workload_connectors.base.BaseConnector.submit`.
        """
        limits = data.get_resource_limits()
        logger.debug(
            __(
                "Connector '{}' running for Data with id {} ({}).",
                self.__class__.__module__,
                data.id,
                repr(argv),
            )
        )

        # Compute target partition.
        partition = getattr(settings, "FLOW_SLURM_PARTITION_DEFAULT", None)
        if data.process.slug in getattr(settings, "FLOW_SLURM_PARTITION_OVERRIDES", {}):
            partition = settings.FLOW_SLURM_PARTITION_OVERRIDES[data.process.slug]

        try:
            # Make sure the resulting file is executable on creation.
            runtime_dir = storage_settings.FLOW_VOLUMES["runtime"]["config"]["path"]
            script_path = os.path.join(runtime_dir, "slurm-{}.sh".format(data.pk))
            file_descriptor = os.open(script_path, os.O_WRONLY | os.O_CREAT, mode=0o555)
            with os.fdopen(file_descriptor, "wt") as script:
                script.write("#!/bin/bash\n")
                script.write(
                    "#SBATCH --mem={}M\n".format(
                        limits["memory"] + EXECUTOR_MEMORY_OVERHEAD
                    )
                )
                script.write("#SBATCH --cpus-per-task={}\n".format(limits["cores"]))
                if partition:
                    script.write("#SBATCH --partition={}\n".format(partition))
                    script.write(
                        "#SBATCH --output slurm-url-{}-job-%j.out\n".format(
                            data.location.subpath
                        )
                    )

                # Render the argument vector into a command line.
                line = " ".join(map(shlex.quote, argv))
                script.write(line + "\n")

            command = ["/usr/bin/env", "sbatch", script_path]
            subprocess.Popen(command, cwd=runtime_dir, stdin=subprocess.DEVNULL).wait()
        except OSError as err:
            logger.error(
                __(
                    "OSError occurred while preparing SLURM script for Data {}: {}",
                    data.id,
                    err,
                )
            )



# Generated by Django 3.1.7 on 2021-10-12 10:39
import os

from django.db import connection, migrations


def load_triggers(apps, schema_editor):
    file_names = ["triggers_entity.sql"]
    with connection.cursor() as c:
        for file_name in file_names:
            file_path = os.path.join(os.path.dirname(__file__), file_name)
            with open(file_path) as fh:
                sql_statement = fh.read()
            c.execute(sql_statement)


class Migration(migrations.Migration):
    dependencies = [("flow", "0023_remove_entity_descriptor_and_more")]

    operations = [migrations.RunPython(load_triggers)]



"""Entity viewset."""

import re
from typing import Optional

from django.db.models import F, Func, OuterRef, Prefetch, Subquery
from django.db.models.functions import Coalesce
from drf_spectacular.utils import extend_schema
from rest_framework import exceptions, serializers, status
from rest_framework.decorators import action
from rest_framework.request import Request
from rest_framework.response import Response

from resolwe.flow.filters import EntityFilter
from resolwe.flow.models import AnnotationValue, Data, Entity
from resolwe.flow.serializers import EntitySerializer
from resolwe.flow.serializers.annotations import AnnotationsByPathSerializer
from resolwe.observers.mixins import ObservableMixin
from resolwe.observers.views import BackgroundTaskSerializer

from .collection import BaseCollectionViewSet
from .utils import get_collection_for_user


class MoveEntityToCollectionSerializer(serializers.Serializer):
    """Serializer for moving entities to a collection."""

    ids = serializers.ListField(child=serializers.IntegerField())
    destination_collection = serializers.IntegerField()


class EntityViewSet(ObservableMixin, BaseCollectionViewSet):
    """API view for entities."""

    serializer_class = EntitySerializer
    filterset_class = EntityFilter

    data_count_subquery = (
        Data.objects.filter(entity_id=OuterRef("pk"))
        .annotate(count=Func(F("id"), function="Count"))
        .values("count")
    )
    data_status_subquery = (
        Data.objects.filter(entity_id=OuterRef("pk"))
        .annotate(
            count=Coalesce(
                Func(
                    F("status"),
                    function="ARRAY_AGG",
                    template="%(function)s(DISTINCT %(expressions)s)",
                ),
                [],
            )
        )
        .values("count")
    )
    queryset = (
        Entity.objects.select_related("contributor")
        .prefetch_related(
            Prefetch("collection", queryset=BaseCollectionViewSet.queryset)
        )
        .annotate(data_statuses=Subquery(data_status_subquery))
        .annotate(data_count=Subquery(data_count_subquery))
    )

    def order_queryset(self, queryset):
        """Order queryset by annotation values.

        The format of the ordering field is annotations__field_id, possibly prefixed by
        a minus sign. The minus sign indicates descending order.
        """
        if ordering := self.request.query_params.get("ordering"):
            order_by = []
            regex = re.compile(r"-?annotations__(?P<field_id>\d+)")
            fields = [field.strip() for field in ordering.split(",")]
            for match in filter(None, map(regex.match, fields)):
                field_id = match.group("field_id")
                # Always filter by user-visible attribute annotation_label.
                annotation_value = AnnotationValue.objects.filter(
                    entity_id=OuterRef("pk"), field_id=field_id
                ).values("_value__label")
                annotate = {f"_order_{field_id}": Subquery(annotation_value)}
                queryset = queryset.annotate(**annotate)
                sign = "-" if match.string.startswith("-") else ""
                order_by.append(f"{sign}_order_{field_id}")
            if order_by:
                queryset = queryset.order_by(*order_by)
        return queryset

    def filter_queryset(self, queryset):
        """Filter queryset."""
        # First apply the default filters.
        queryset = super().filter_queryset(queryset)
        # Order by annotation values if required.
        return self.order_queryset(queryset)

    def _get_entities(self, user, ids):
        """Return entities queryset based on provided entity ids."""
        queryset = Entity.objects.filter(id__in=ids).filter_for_user(user)
        actual_ids = queryset.values_list("id", flat=True)
        missing_ids = list(set(ids) - set(actual_ids))
        if missing_ids:
            raise exceptions.ParseError(
                "Entities with the following ids not found: {}".format(
                    ", ".join(map(str, missing_ids))
                )
            )

        return queryset

    @extend_schema(
        request=MoveEntityToCollectionSerializer(), responses={status.HTTP_200_OK: None}
    )
    @action(detail=False, methods=["post"])
    def move_to_collection(self, request, *args, **kwargs):
        """Move samples from source to destination collection."""

        serializer = MoveEntityToCollectionSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        ids = serializer.validated_data["ids"]
        dst_collection_id = serializer.validated_data["destination_collection"]
        dst_collection = get_collection_for_user(dst_collection_id, request.user)
        entity_qs = self._get_entities(request.user, ids)
        task = entity_qs.move_to_collection(dst_collection, request.user)
        return Response(
            status=status.HTTP_200_OK,
            data=BackgroundTaskSerializer(task).data,
        )

    # NOTE: This can be deleted when DRF will support select_for_update
    #       on updates and ResolweUpdateModelMixin will use it.
    #       https://github.com/encode/django-rest-framework/issues/4675
    def update(self, request, *args, **kwargs):
        """Update an entity.

        Original queryset produces a temporary database table whose rows
        cannot be selected for an update. As a workaround, we patch
        get_queryset function to return only Entity objects without
        additional data that is not needed for the update.
        """
        orig_get_queryset = self.get_queryset

        def patched_get_queryset():
            """Patched get_queryset method."""
            entity_ids = orig_get_queryset().values_list("id", flat=True)
            return Entity.objects.filter(id__in=entity_ids)

        self.get_queryset = patched_get_queryset

        resp = super().update(request, *args, **kwargs)

        self.get_queryset = orig_get_queryset
        return resp

    @extend_schema(
        request=AnnotationsByPathSerializer(many=True),
        responses={status.HTTP_200_OK: None},
    )
    @action(detail=True, methods=["post"])
    def set_annotations(self, request: Request, pk: Optional[int] = None):
        """Add the given list of AnnotationFields to the given entity."""
        # No need to check for permissions, since post requires edit by default.
        entity = self.get_object()
        # Read and validate the request data.
        serializer = AnnotationsByPathSerializer(data=request.data, many=True)
        serializer.is_valid(raise_exception=True)
        annotations = {value["field_path"]: value["value"] for value in serializer.data}
        entity.update_annotations(annotations, request.user)
        return Response()



# pylint: disable=missing-docstring

from resolwe.flow.models import Collection, DescriptorSchema, Process
from resolwe.permissions.models import Permission
from resolwe.permissions.utils import copy_permissions
from resolwe.test import TestCase


class UtilsTestCase(TestCase):
    def setUp(self):
        super().setUp()

        self.src_process = Process.objects.create(
            name="Source process", contributor=self.contributor
        )
        self.dst_process = Process.objects.create(
            name="Destination process", contributor=self.contributor
        )

        self.collection = Collection.objects.create(
            name="Test collection", contributor=self.contributor
        )

        self.descriptor_schema = DescriptorSchema.objects.create(
            contributor=self.contributor
        )

    def test_copy_permissions(self):
        self.src_process.set_permission(Permission.VIEW, self.contributor)
        self.src_process.set_permission(Permission.VIEW, self.group)

        copy_permissions(self.src_process, self.dst_process)
        self.assertTrue(self.contributor.has_perm(Permission.VIEW, self.dst_process))
        # User inherites permission from group
        self.assertTrue(self.user.has_perm(Permission.VIEW, self.dst_process))
